from typing import Optional


def replace_search_term_enclosures(query: str, open_replacement: str, close_replacement: str,
                                   only_on_wildcards: Optional[bool] = False) -> str:
    """
    Replace search term enclosures

    Parameters
    ----------
    query : str
        A search query
    open_replacement : str
        An open replacement string
    close_replacement : str
        An open replacement string
    only_on_wildcards : bool, optional
        If the replacement should be only be performed on wildcards terms, by default False

    Returns
    -------
    str
        A transformed query
    """

    if only_on_wildcards:

        is_inside_a_term = False
        search_term = ''
        final_query = ''
        for character in query:

            if character == '[':
                search_term += character
                is_inside_a_term = True
                continue

            if is_inside_a_term:
                search_term += character
                if character == ']':
                    if '?' in search_term or '*' in search_term:
                        search_term = search_term.replace(
                            '[', open_replacement).replace(']', close_replacement)
                    final_query += search_term
                    search_term = ''
                    is_inside_a_term = False
            else:
                final_query += character

        return final_query

    else:

        return query.replace('[', open_replacement).replace(']', close_replacement)



import os
import random
import requests
import findpapers.utils.common_util as common_util


# list of most common user agents (Last Updated: Wed, 09 Sep 2020)
USER_AGENTS = [
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.2 Safari/605.1.15',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.1 Safari/605.1.15',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; rv:78.0) Gecko/20100101 Firefox/78.0',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.2 Safari/605.1.15',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:80.0) Gecko/20100101 Firefox/80.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:80.0) Gecko/20100101 Firefox/80.0',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36 Edg/85.0.564.44',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:78.0) Gecko/20100101 Firefox/78.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; rv:68.0) Gecko/20100101 Firefox/68.0',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:68.0) Gecko/20100101 Firefox/68.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36 Edg/84.0.522.63',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.2 Safari/605.1.15',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1 Safari/605.1.15',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36 OPR/70.0.3728.106',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36 Edg/84.0.522.59',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36 Edg/84.0.522.52',
    'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36 Edg/85.0.564.41',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36 Edge/18.18362',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36 Edge/18.18363',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 11_0) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Safari/605.1.15',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36 OPR/69.0.3686.95',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.83 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36 Edg/84.0.522.61',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:80.0) Gecko/20100101 Firefox/80.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.135 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:77.0) Gecko/20100101 Firefox/77.0',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64; rv:79.0) Gecko/20100101 Firefox/79.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.138 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:68.0) Gecko/20100101 Firefox/68.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36 Edg/84.0.522.58',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.86 YaBrowser/20.8.0.894 Yowser/2.5 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.2 Safari/605.1.15',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.89 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.1.2 Safari/605.1.15',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36',
]


class DefaultSession(requests.Session, metaclass=common_util.ThreadSafeSingletonMetaclass):

    """
    Session class with singleton feature and custom headers config
    """

    def __init__(self, *args, **kwargs):

        super(DefaultSession, self).__init__()

        PROXY = os.getenv('FINDPAPERS_PROXY')

        if PROXY is not None:

            self.proxies = {
                'http': PROXY,
                'https': PROXY
            }

        self.headers.update({'User-Agent': random.choice(USER_AGENTS)})

    def request(self, method, url, **kwargs):
        """
        This is just a common request, the only difference is that when proxies are provided
        and a response isn't ok, we'll try one more time without using the proxies
        """

        try:
            response = super().request(method, url, **kwargs)
        except Exception:
            response = requests.Response()
            response.status_code = 500

        if not response.ok and ('http' in self.proxies or 'https' in self.proxies):
            # if the response is not ok using proxies,
            # let's try one more time without using them
            kwargs['proxies'] = {
                'http': None,
                'https': None,
            }
            response = super().request(method, url, **kwargs)

        return response



from __future__ import annotations
import datetime
import itertools
import edlib
from typing import List, Optional
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication


class Search():
    """
    Class that represents a search
    """

    def __init__(self, query: str, since: Optional[datetime.date] = None, until: Optional[datetime.date] = None,
                 limit: Optional[int] = None, limit_per_database: Optional[int] = None, processed_at: Optional[datetime.datetime] = None,
                 databases: Optional[List[str]] = None, publication_types: Optional[List[str]] = None, papers: Optional[set] = None):
        """
        Class constructor

        Parameters
        ----------
        query : str
            The query used to fetch the papers
        since : datetime.date, optional
            The lower bound (inclusive) date of search, by default None
        until : datetime.date, optional
            The upper bound (inclusive) date of search, by default None
        limit : int, optional
            The max number of papers that can be returned in the search, 
            when the limit is not provided the search will retrieve all the papers that it can, by default None
        limit_per_database : int, optional
            The max number of papers that can be returned in the search for each database
            when the limit is not provided the search will retrieve all the papers that it can, by default None
        processed_at : datetime.datetime, optional
            The datetime when the search was performed
        databases : List[str], optional
            List of databases where the search should be performed, if not specified all databases will be used. by default None
        publication_types : List[str], optional
            List of publication list of publication types to filter when searching, if not specified all publication types will be used. By default None
        papers : set, optional
            A list of papers already collected
        """

        self.query = query
        self.since = since
        self.until = until
        self.limit = limit
        self.limit_per_database = limit_per_database
        self.processed_at = processed_at if processed_at is not None else datetime.datetime.utcnow()
        self.databases = databases
        self.publication_types = publication_types

        self.paper_by_key = {}
        self.publication_by_key = {}
        self.paper_by_doi = {}
        self.papers_by_database = {}

        self.papers = set()
        if papers is not None:
            for paper in papers:
                try:
                    self.add_paper(paper)
                except Exception:
                    pass

    def get_paper_key(self, paper_title: str, publication_date: datetime.date, paper_doi: Optional[str] = None) -> str:
        """
        We have a map called paper_by_key that is filled using the string this method returns

        Parameters
        ----------
        paper_title : str
            The paper title
        publication_date : datetime.date
            The paper publication date
        paper_doi : str, optional
            The paper DOI, by default None

        Returns
        -------
        str
            A string that represents a unique key for each paper, that will be used to fill and retrieve values from paper_by_key
        """

        if paper_doi is not None:
            return f'DOI-{paper_doi}'
        else:
            return f'{paper_title.lower()}|{publication_date.year if publication_date is not None else ""}'

    def get_publication_key(self, publication_title: str, publication_issn: Optional[str] = None, publication_isbn: Optional[str] = None) -> str:
        """
        We have a map called publication_by_key that is filled using the string this method returns

        Parameters
        ----------
        publication_title : str
            The publication title
        publication_issn : Optional[str], optional
            The publication issn, by default None
        publication_isbn : Optional[str], optional
            The publication isbn, by default None

        Returns
        -------
        str
            A string that represents a unique key for each publication, that will be used to fill and retrieve values from publication_by_key
        """

        if publication_issn is not None:
            return f'ISSN-{publication_issn.lower()}'
        elif publication_isbn is not None:
            return f'ISBN-{publication_isbn.lower()}'
        else:
            return f'TITLE-{publication_title.lower()}'

    def add_paper(self, paper: Paper):
        """
        Method that handle the action to add a paper to the list of already collected papers, 
        dealing with possible paper's duplications

        Parameters
        ----------
        paper : Paper
            A new collected paper instance
        Raises
        ------
        ValueError
            - Paper cannot be added to search without at least one defined database
        OverflowError
            - When the papers limit is provided, you cannot exceed it
        """

        if len(paper.databases) == 0:
            raise ValueError(
                'Paper cannot be added to search without at least one defined database')

        for database in paper.databases:
            if self.databases is not None and database.lower() not in self.databases:
                raise ValueError(f'Database {database} isn\'t in databases list')
            if self.reached_its_limit(database):
                raise OverflowError('When the papers limit is provided, you cannot exceed it')

        if database not in self.papers_by_database:
            self.papers_by_database[database] = set()

        if paper.publication is not None:

            publication_key = self.get_publication_key(
                paper.publication.title, paper.publication.issn, paper.publication.isbn)
            already_collected_publication = self.publication_by_key.get(
                publication_key, None)

            if already_collected_publication is not None:
                already_collected_publication.enrich(paper.publication)
                paper.publication = already_collected_publication
            else:
                self.publication_by_key[publication_key] = paper.publication

        paper_key = self.get_paper_key(
            paper.title, paper.publication_date, paper.doi)

        already_collected_paper = self.paper_by_key.get(paper_key, None)

        if (self.since is None or paper.publication_date >= self.since) \
                and (self.until is None or paper.publication_date <= self.until):

            if already_collected_paper is None:
                self.papers.add(paper)
                self.paper_by_key[paper_key] = paper

                if paper.doi is not None:
                    self.paper_by_doi[paper.doi] = paper

                for database in paper.databases:
                    if database not in self.papers_by_database:
                        self.papers_by_database[database] = set()
                    self.papers_by_database[database].add(paper)
            else:
                self.papers_by_database[database].add(already_collected_paper)
                already_collected_paper.enrich(paper)
        

    def get_paper(self, paper_title: str, publication_date: str, paper_doi: Optional[str] = None) -> Paper:
        """
        Get a collected paper by paper's title and publication date

        Parameters
        ----------
        paper_title : str
            The paper title
        publication_date : datetime.date
            The paper publication date
        paper_doi : str, optional
            The paper DOI, by default None

        Returns
        -------
        Paper
            The wanted paper, or None if there isn't a paper given by the provided arguments
        """

        paper_key = self.get_paper_key(
            paper_title, publication_date, paper_doi)

        return self.paper_by_key.get(paper_key, None)

    def get_publication(self, title: str, issn: Optional[str] = None, isbn: Optional[str] = None) -> Publication:
        """
        Get a collected publication by publication's title, issn and isbn

        Parameters
        ----------
        title : str
            The publication title
        issn : Optional[str]
            The publication ISSN, by default None
        isbn : Optional[str]
            The publication ISBN, by default None

        Returns
        -------
        Publication
            The wanted publication, or None if there isn't a publication given by the provided arguments
        """

        publication_key = self.get_publication_key(title, issn, isbn)

        return self.publication_by_key.get(publication_key, None)

    def remove_paper(self, paper: Paper):
        """
        Remove a collected paper

        Parameters
        ----------
        paper : Paper
            A paper instance
        """

        paper_key = self.get_paper_key(
            paper.title, paper.publication_date, paper.doi)

        if paper_key in self.paper_by_key:
            del self.paper_by_key[paper_key]

        for database in paper.databases:
            self.papers_by_database[database].remove(paper)

        self.papers.remove(paper)

    def merge_duplications(self, similarity_threshold: float = 0.95):
        """
        In some cases, a same paper is represented with tiny differences between some databases, 
        this method try to deal with this situation merging those instances of the paper,
        using a similarity threshold, by default 0.95 (95%), i.e., if two papers titles
        are similar by 95% or more, and if the papers have the same year of publication
        this papers are considered duplications of a same paper.

        Parameters
        ----------
        max_similarity_threshold : float, optional
            A value between 0 and 1 that represents a threshold that says if a pair of papers is a duplication or not, by default 0.95 (95%)
        """

        paper_key_pairs = list(
            itertools.combinations(self.paper_by_key.keys(), 2))

        for i, pair in enumerate(paper_key_pairs):

            paper_1_key = pair[0]
            paper_2_key = pair[1]
            paper_1 = self.paper_by_key.get(paper_1_key)
            paper_2 = self.paper_by_key.get(paper_2_key)

            if (paper_1.publication_date is None or paper_2.publication_date is None) or \
                (paper_1.publication_date.year != paper_2.publication_date.year) or \
                (paper_1.doi is not None and paper_2.doi is not None and paper_1.doi != paper_2.doi):
                # We cannot merge paper from different years or without a year defined or different DOI
                break

            max_title_length = max(len(paper_1.title), len(paper_2.title))

            # creating the max valid edit distance using the max title length between the two papers and the provided similarity threshold
            max_edit_distance = int(
                max_title_length * (1 - similarity_threshold))

            # calculating the edit distance between the titles
            titles_edit_distance = edlib.align(
                paper_1.title.lower(), paper_2.title.lower())['editDistance']

            if (paper_1.doi is not None and paper_1.doi == paper_2.doi) or (titles_edit_distance <= max_edit_distance):

                # using the information of paper_2 to enrich paper_1
                paper_1.enrich(paper_2)

                # removing the paper_2 instance
                self.remove_paper(paper_2)

    def reached_its_limit(self, database: str) -> bool:
        """
        Returns a flag that says if the search has reached its limit

        Parameters
        ----------
        database : str, optional
            The database name that will be used to check the limit

        Returns
        -------
        bool
            a flag that says if the search has reached its limit
        """

        reached_general_limit = self.limit is not None and len(
            self.papers) >= self.limit
        reached_database_limit = self.limit_per_database is not None and database in self.papers_by_database and len(
            self.papers_by_database.get(database)) >= self.limit_per_database

        return reached_general_limit or reached_database_limit

    @classmethod
    def from_dict(cls, search_dict: dict) -> Search:
        """
        A method that returns a Search instance based on the provided dict object

        Parameters
        ----------
        search_dict : dict
            A dict that represents a Search instance

        Returns
        -------
        Search
            A Search instance based on the provided dict object
        """

        query = search_dict.get('query')
        limit = search_dict.get('limit')
        limit_per_database = search_dict.get('limit_per_database')

        since = search_dict.get('since')
        if since is not None:
            since = datetime.datetime.strptime(since, '%Y-%m-%d').date()

        until = search_dict.get('until')
        if until is not None:
            until = datetime.datetime.strptime(until, '%Y-%m-%d').date()

        processed_at = search_dict.get('processed_at')
        if processed_at is not None:
            processed_at = datetime.datetime.strptime(processed_at, '%Y-%m-%d %H:%M:%S')

        databases = search_dict.get('databases')
        publication_types = search_dict.get('publication_types')

        papers = set()
        for paper in search_dict.get('papers', []):
            papers.add(Paper.from_dict(paper))

        return cls(query, since, until, limit, limit_per_database, processed_at, databases, publication_types, papers)

    @staticmethod
    def to_dict(search: Search) -> dict:
        """
        A method that returns a dict object based on the provided Search instance

        Parameters
        ----------
        search : Search
            A Search instance

        Returns
        -------
        dict
            A dict that represents a Search instance
        """

        papers = []
        for paper in search.papers:
            papers.append(Paper.to_dict(paper))

        papers = sorted(papers, key=lambda x: x.get('publication_date', '1900'), reverse=True)

        number_of_papers_by_database = {}
        for database, items in search.papers_by_database.items():
            number_of_papers_by_database[database] = len(items)

        return {
            'query': search.query,
            'since': search.since.strftime('%Y-%m-%d') if search.since is not None else None,
            'until': search.until.strftime('%Y-%m-%d') if search.until is not None else None,
            'limit': search.limit,
            'limit_per_database': search.limit_per_database,
            'processed_at': search.processed_at.strftime('%Y-%m-%d %H:%M:%S') if search.processed_at is not None else None,
            'databases': search.databases,
            'publication_types': search.publication_types,
            'number_of_papers': len(papers),
            'number_of_papers_by_database': number_of_papers_by_database,
            'papers': papers
        }



import os
import re
import json
import requests
import logging
import datetime
import urllib.parse
from lxml import html
from typing import Optional, List
import findpapers.utils.common_util as common_util
import findpapers.utils.persistence_util as persistence_util
from findpapers.models.search import Search
from findpapers.utils.requests_util import DefaultSession


def download(search_path: str, output_directory: str, only_selected_papers: Optional[bool] = False,
             categories_filter: Optional[dict] = None, proxy: Optional[str] = None):
    """
    If you've done your search, (probably made the search refinement too) and wanna download the papers, 
    this is the method that you need to call. This method will try to download the PDF version of the papers to
    the output directory path.

    We use some heuristics to do our job, but sometime they won't work properly, and we cannot be able
    to download the papers, but we logging the downloads or failures in a file download.log
    placed on the output directory, you can check out the log to find what papers cannot be downloaded
    and try to get them manually later. 

    Note: Some papers are behind a paywall and won't be able to be downloaded by this method. 
    However, if you have a proxy provided for the institution where you study or work that permit you 
    to "break" this paywall. You can use this proxy configuration here
    by setting the environment variables FINDPAPERS_HTTP_PROXY and FINDPAPERS_HTTPS_PROXY.

    Parameters
    ----------
    search_path : str
        A valid file path containing a JSON representation of the search results
    output_directory : str
        A valid file path of the directory where the downloaded papers will be placed
    only_selected_papers : bool, False by default
        If only the selected papers will be downloaded
    categories_filter : dict, None by default
        A dict of categories to be used to filter which papers will be downloaded
    proxy : Optional[str], optional
        proxy URL that can be used during requests. This can be also defined by an environment variable FINDPAPERS_PROXY. By default None
    """

    if proxy is not None:
        os.environ['FINDPAPERS_PROXY'] = proxy

    search = persistence_util.load(search_path)

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    log_filepath = os.path.join(output_directory, 'download.log')

    common_util.check_write_access(log_filepath)

    with open(log_filepath, 'a' if os.path.exists(log_filepath) else 'w') as fp:
        now = datetime.datetime.now()
        fp.write(
            f"------- A new download process started at: {datetime.datetime.strftime(now, '%Y-%m-%d %H:%M:%S')} \n")

    for i, paper in enumerate(search.papers):

        logging.info(f'({i+1}/{len(search.papers)}) {paper.title}')

        if (only_selected_papers and not paper.selected) or \
        (categories_filter is not None and (paper.categories is None or not paper.has_category_match(categories_filter))):
            continue

        downloaded = False
        output_filename = f'{paper.publication_date.year}-{paper.title}'
        output_filename = re.sub(
            r'[^\w\d-]', '_', output_filename)  # sanitize filename
        output_filename += '.pdf'
        output_filepath = os.path.join(output_directory, output_filename)

        if os.path.exists(output_filepath):  # PDF already collected
            logging.info(f'Paper\'s PDF file has already been collected')
            continue

        if paper.doi is not None:
            paper.urls.add(f'http://doi.org/{paper.doi}')

        for url in paper.urls:  # we'll try to download the PDF file of the paper by its URLs
            try:
                logging.info(f'Fetching data from: {url}')

                response = common_util.try_success(
                    lambda url=url: DefaultSession().get(url), 2)

                if response is None:
                    continue

                if 'text/html' in response.headers.get('content-type').lower():

                    response_url = urllib.parse.urlsplit(response.url)
                    response_query_string = urllib.parse.parse_qs(
                        urllib.parse.urlparse(response.url).query)
                    response_url_path = response_url.path
                    host_url = f'{response_url.scheme}://{response_url.hostname}'
                    pdf_url = None

                    if response_url_path.endswith('/'):
                        response_url_path = response_url_path[:-1]

                    response_url_path = response_url_path.split('?')[0]

                    if host_url in ['https://dl.acm.org']:

                        doi = paper.doi
                        if doi is None and response_url_path.startswith('/doi/') and '/doi/pdf/' not in response_url_path:
                            doi = response_url_path[4:]
                        elif doi is None:
                            continue

                        pdf_url = f'https://dl.acm.org/doi/pdf/{doi}'

                    elif host_url in ['https://ieeexplore.ieee.org']:

                        if response_url_path.startswith('/document/'):
                            document_id = response_url_path[10:]
                        elif response_query_string.get('arnumber', None) is not None:
                            document_id = response_query_string.get('arnumber')[
                                0]
                        else:
                            continue

                        pdf_url = f'{host_url}/stampPDF/getPDF.jsp?tp=&arnumber={document_id}'

                    elif host_url in ['https://www.sciencedirect.com', 'https://linkinghub.elsevier.com']:

                        paper_id = response_url_path.split('/')[-1]
                        pdf_url = f'https://www.sciencedirect.com/science/article/pii/{paper_id}/pdfft?isDTMRedir=true&download=true'

                    elif host_url in ['https://pubs.rsc.org']:

                        pdf_url = response.url.replace(
                            '/articlelanding/', '/articlepdf/')

                    elif host_url in ['https://www.tandfonline.com', 'https://www.frontiersin.org']:

                        pdf_url = response.url.replace('/full', '/pdf')

                    elif host_url in ['https://pubs.acs.org', 'https://journals.sagepub.com', 'https://royalsocietypublishing.org']:

                        pdf_url = response.url.replace('/doi', '/doi/pdf')

                    elif host_url in ['https://link.springer.com']:

                        pdf_url = response.url.replace(
                            '/article/', '/content/pdf/').replace('%2F', '/') + '.pdf'

                    elif host_url in ['https://www.isca-speech.org']:

                        pdf_url = response.url.replace(
                            '/abstracts/', '/pdfs/').replace('.html', '.pdf')

                    elif host_url in ['https://onlinelibrary.wiley.com']:

                        pdf_url = response.url.replace(
                            '/full/', '/pdfdirect/').replace('/abs/', '/pdfdirect/')

                    elif host_url in ['https://www.jmir.org', 'https://www.mdpi.com']:

                        pdf_url = response.url + '/pdf'

                    elif host_url in ['https://www.pnas.org']:

                        pdf_url = response.url.replace(
                            '/content/', '/content/pnas/') + '.full.pdf'

                    elif host_url in ['https://www.jneurosci.org']:

                        pdf_url = response.url.replace(
                            '/content/', '/content/jneuro/') + '.full.pdf'

                    elif host_url in ['https://www.ijcai.org']:

                        paper_id = response.url.split('/')[-1].zfill(4)
                        pdf_url = '/'.join(response.url.split('/')
                                           [:-1]) + '/' + paper_id + '.pdf'

                    elif host_url in ['https://asmp-eurasipjournals.springeropen.com']:

                        pdf_url = response.url.replace(
                            '/articles/', '/track/pdf/')

                    if pdf_url is not None:

                        response = common_util.try_success(
                            lambda url=pdf_url: DefaultSession().get(url), 2)

                if 'application/pdf' in response.headers.get('content-type').lower():
                    with open(output_filepath, 'wb') as fp:
                        fp.write(response.content)
                    downloaded = True
                    break

            except Exception as e:  # pragma: no cover
                logging.debug(e, exc_info=True)

        if downloaded:
            with open(log_filepath, 'a') as fp:
                fp.write(f'[DOWNLOADED] {paper.title}\n')
        else:
            with open(log_filepath, 'a') as fp:
                fp.write(f'[FAILED] {paper.title}\n')
                if len(paper.urls) == 0:
                    fp.write(f'Empty URL list\n')
                else:
                    for url in paper.urls:
                        fp.write(f'{url}\n')



import logging
from typing import Optional
import findpapers.utils.persistence_util as persistence_util
import findpapers.utils.common_util as common_util


def generate_bibtex(search_path: str, outputpath: str, only_selected_papers: Optional[bool] = False,
                    categories_filter: Optional[dict] = None):
    """
    Method used to generate a BibTeX file from a search result

    Parameters
    ----------
    search_path : str
        A valid file path containing a JSON representation of the search results
    outputpath : str
        A valid file path for the BibTeX output file
    only_selected_papers : bool, optional
        If you only want to generate a BibTeX file for selected papers, by default False
    categories_filter : dict, None by default
        A dict of categories to be used to filter which papers will be downloaded
    """

    search = persistence_util.load(search_path)
    common_util.check_write_access(outputpath)

    default_tab = ' ' * 4
    bibtex_output = ''

    for paper in search.papers:

        if (only_selected_papers and not paper.selected) or \
        (categories_filter is not None and (paper.categories is None or not paper.has_category_match(categories_filter))):
            continue

        logging.info(f'Exporting bibtex for: {paper.title}')

        try:

            citation_type = '@unpublished'
            if paper.publication is not None:
                if paper.publication.category == 'Journal':
                    citation_type = '@article'
                elif paper.publication.category == 'Conference Proceedings':
                    citation_type = '@inproceedings'
                elif paper.publication.category == 'Book':
                    citation_type = '@book'
                else:
                    citation_type = '@misc'

            bibtex_output += f'{citation_type}{"{"}{paper.get_citation_key()},\n'

            bibtex_output += f'{default_tab}title = {{{paper.title}}},\n'

            if len(paper.authors) > 0:
                authors = ' and '.join(paper.authors)
                bibtex_output += f'{default_tab}author = {{{authors}}},\n'

            if citation_type == '@unpublished':
                note = ''
                if len(paper.urls) > 0:
                    note += f'Available at {list(paper.urls)[0]}'
                if paper.publication_date is not None:
                    note += f' ({paper.publication_date.strftime("%Y/%m/%d")})'
                if paper.comments is not None:
                    note += paper.comments if len(
                        note) == 0 else f' | {paper.comments}'
                bibtex_output += f'{default_tab}note = {{{note}}},\n'
            elif citation_type == '@article':
                bibtex_output += f'{default_tab}journal = {{{paper.publication.title}}},\n'
            elif citation_type == '@inproceedings':
                bibtex_output += f'{default_tab}booktitle = {{{paper.publication.title}}},\n'
            elif citation_type == '@misc' and len(paper.urls) > 0 and paper.publication_date is not None:
                date = paper.publication_date.strftime('%Y/%m/%d')
                url = list(paper.urls)[0]
                bibtex_output += f'{default_tab}howpublished = {{Available at {url} ({date})}},\n'

            if paper.publication is not None and paper.publication.publisher is not None:
                bibtex_output += f'{default_tab}publisher = {{{paper.publication.publisher}}},\n'

            if paper.publication_date is not None:
                bibtex_output += f'{default_tab}year = {{{paper.publication_date.year}}},\n'

            if paper.pages is not None:
                bibtex_output += f'{default_tab}pages = {{{paper.pages}}},\n'

            bibtex_output = bibtex_output.rstrip(
                ',\n') + '\n'  # removing last comma

            bibtex_output += '}\n\n'

        except Exception as e:
            logging.debug(e, exc_info=True)

    with open(outputpath, 'w') as fp:
        fp.write(bibtex_output)



from __future__ import annotations
from typing import List, Optional
import json


class Publication():
    """
    Class that represents a publication (journal, conference proceedings, book) instance
    """

    def __init__(self, title: str, isbn: Optional[str] = None, issn: Optional[str] = None, publisher: Optional[str] = None,
                 category: Optional[str] = None, cite_score: Optional[float] = None, sjr: Optional[float] = None,
                 snip: Optional[float] = None, subject_areas: Optional[set] = None):
        """
        Paper class constructor

        Parameters
        ----------
        title : str
            publication title
        isbn : str, optional
            publication ISBN, by default None
        issn : str, optional
            publication ISSN, by default None
        publisher : str, optional
            publication publisher, by default None
        category : str, optional
            publication category (Journal, Conference Proceedings, Book), by default None
        cite_score : float, optional
            CiteScore measures average citations received per document published in the serial, by default None
        sjr : float, optional
            SCImago Journal Rank measures weighted citations received by the serial. 
            Citation weighting depends on subject field and prestige (SJR) of the citing serial, by default None
        snip : float, optional
            Source Normalized Impact per Paper measures actual citations received relative to citations 
            expected for the serial’s subject field, by default None
        subject_areas : float, optional
            Publication subjects areas, by default None
        Raises
        ------
        ValueError
            - Publication's title cannot be null
        """

        if title is None or len(title) == 0:
            raise(ValueError('Publication\'s title cannot be null'))

        self.title = title
        self.isbn = isbn
        self.issn = issn
        self.publisher = publisher
        self.category = category if category is not None else title # trying to figure out what is the category by publication title
        self.cite_score = cite_score
        self.sjr = sjr
        self.snip = snip
        self.subject_areas = subject_areas if subject_areas is not None else set()

    @property
    def category(self):
        return self._category

    @category.setter
    def category(self, value: str):
        """
        Category value setter, this method also try to convert a provided invalid category 
        to a valid one [Journal, Conference Proceedings, Book]

        Parameters
        ----------
        value : str
            A category string
        """

        if value is not None:

            # trying to convert a provided invalid category to a valid one [Journal, Conference Proceedings, Book]
            if 'journal' in value.lower():
                value = 'Journal'
            elif 'conference' in value.lower() or 'proceeding' in value.lower():
                value = 'Conference Proceedings'
            elif 'book' in value.lower():
                value = 'Book'
            else:
                value = None

        self._category = value

    def enrich(self, publication: Publication):
        """
        e can enrich some publication information using a duplication of it found in another database.
        This method do this using a provided instance of a duplicated publication

        Parameters
        ----------
        publication : Publication
            A duplication of the "self" publication
        """

        if self.title is None or (publication.title is not None and len(self.title) < len(publication.title)):
            self.title = publication.title

        if self.isbn is None:
            self.isbn = publication.isbn

        if self.issn is None:
            self.issn = publication.issn

        if self.publisher is None:
            self.publisher = publication.publisher

        if self.category is None and publication.category is not None:
            self.category = publication.category

        if self.cite_score is None:
            self.cite_score = publication.cite_score

        if self.sjr is None:
            self.sjr = publication.sjr

        if self.snip is None:
            self.snip = publication.snip

        for subject_area in publication.subject_areas:
            if subject_area is not None and len(subject_area.strip()) > 0:
                self.subject_areas.add(subject_area.strip())

    @classmethod
    def from_dict(cls, publication_dict: dict) -> Publication:
        """
        A method that returns a Publication instance based on the provided dict object

        Parameters
        ----------
        publication_dict : dict
            A dict that represents a Publication instance

        Returns
        -------
        Publication
            A Publication instance based on the provided dict object
        """

        title = publication_dict.get('title')
        isbn = publication_dict.get('isbn')
        issn = publication_dict.get('issn')
        publisher = publication_dict.get('publisher')
        category = publication_dict.get('category')
        cite_score = publication_dict.get('cite_score')
        sjr = publication_dict.get('sjr')
        snip = publication_dict.get('snip')
        subject_areas = set(publication_dict.get('subject_areas'))

        return cls(title, isbn, issn, publisher, category, cite_score, sjr, snip, subject_areas)

    @staticmethod
    def to_dict(publication: Publication) -> dict:
        """
        A method that returns a dict object based on the provided Publication instance

        Parameters
        ----------
        publication : Publication
            A Publication instance

        Returns
        -------
        dict
            A dict that represents a Publication instance
        """

        return {
            'title': publication.title,
            'isbn': publication.isbn,
            'issn': publication.issn,
            'publisher': publication.publisher,
            'category': publication.category,
            'cite_score': publication.cite_score,
            'sjr': publication.sjr,
            'snip': publication.snip,
            'subject_areas': list(publication.subject_areas)
        }



import requests
import datetime
import logging
import re
from lxml import html
from typing import Optional
import findpapers.utils.common_util as common_util
import findpapers.utils.query_util as query_util
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
from findpapers.utils.requests_util import DefaultSession


DATABASE_LABEL = 'Scopus'
BASE_URL = 'https://api.elsevier.com'


def _get_query(search: Search) -> str:
    """
    Get the translated query from search instance to fetch data from Scopus database
    See https://dev.elsevier.com/tips/ScopusSearchTips.htm for query tips

    Parameters
    ----------
    search : Search
        A search instance

    Returns
    -------
    str
        The translated query
    """

    query = query_util.replace_search_term_enclosures(search.query, '"', '"', True)
    query = query_util.replace_search_term_enclosures(query, '{', '}')

    query = f'TITLE-ABS-KEY({query})'

    if search.since is not None:
        query += f' AND PUBYEAR > {search.since.year - 1}'
    if search.until is not None:
        query += f' AND PUBYEAR < {search.until.year + 1}'

    if search.publication_types is not None:

        publication_types = set()

        if 'conference proceedings' in search.publication_types:
            publication_types.add('p') # Conference Proceeding
        if 'journal' in search.publication_types:
            publication_types.add('j') # Journal
        if 'book' in search.publication_types:
            publication_types.add('b') # Book
            publication_types.add('k') # Book Series
        if 'other' in search.publication_types:
            publication_types.add('r') # Report
            publication_types.add('d') # Trade Publication
        
        query += f' AND SRCTYPE({" OR ".join(publication_types)})'

    return query


def _get_publication_entry(publication_issn: str, api_token: str) -> dict:  # pragma: no cover
    """
    Get publication entry by publication ISSN

    Parameters
    ----------
    publication_issn : str
        A publication ISSN
    api_token : str
        A Scopus API token

    Returns
    -------
    dict (or None)
        publication entry in dict format, or None if the API doesn't return a valid entry
    """

    url = f'{BASE_URL}/content/serial/title/issn/{publication_issn}?apiKey={api_token}'
    headers = {'Accept': 'application/json'}
    response = common_util.try_success(lambda: DefaultSession().get(
        url, headers=headers).json().get('serial-metadata-response', None), 2)

    if response is not None and 'entry' in response and len(response.get('entry')) > 0:
        return response.get('entry')[0]


def _get_publication(paper_entry: dict, api_token: str) -> Publication:
    """
    Using a paper entry provided, this method builds a publication instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from scopus API
    api_token : str
        A Scopus API token

    Returns
    -------
    Publication
        A publication instance or None
    """

    # getting data

    publication_title = paper_entry.get('prism:publicationName', None)

    if publication_title is None or len(publication_title) == 0:
        return None

    publication_isbn = paper_entry.get('prism:isbn', None)
    publication_issn = paper_entry.get('prism:issn', None)
    publication_category = paper_entry.get('prism:aggregationType', None)

    if isinstance(publication_isbn, list):
        publication_isbn = publication_isbn[0].get('$')

    if isinstance(publication_issn, list):
        publication_issn = publication_issn[0].get('$')

    publication = Publication(publication_title, publication_isbn,
                              publication_issn, None, publication_category)

    return publication


def _get_paper_page(url: str) -> object:  # pragma: no cover
    """
    Get a paper page element from a provided URL

    Parameters
    ----------
    url : str
        The paper URL

    Returns
    -------
    Object
        A HTML element representing the paper given by the provided URL
    """

    response = common_util.try_success(lambda: DefaultSession().get(url), 2)
    return html.fromstring(response.content)


def _get_paper(paper_entry: dict, publication: Publication) -> Paper:
    """
    Using a paper entry provided, this method builds a paper instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from scopus API
    publication : Publication
        A publication instance that will be associated with the paper

    Returns
    -------
    Paper
        A paper instance or None
    """

    # getting data

    paper_title = paper_entry.get('dc:title', None)

    if paper_title is None or len(paper_title) == 0:
        return None

    paper_publication_date = paper_entry.get('prism:coverDate', None)
    paper_doi = paper_entry.get('prism:doi', None)
    paper_citations = paper_entry.get('citedby-count', None)
    paper_first_author = paper_entry.get('dc:creator', None)
    paper_abstract = None
    paper_authors = []
    paper_urls = set()
    paper_keywords = set()
    paper_pages = None
    paper_number_of_pages = None

    # post processing data

    if paper_first_author is not None:
        paper_authors.append(paper_first_author)

    if paper_publication_date is not None:
        date_split = paper_publication_date.split('-')
        paper_publication_date = datetime.date(
            int(date_split[0]), int(date_split[1]), int(date_split[2]))

    if paper_publication_date is None:
        return None

    if paper_citations is not None:
        paper_citations = int(paper_citations)

    # enriching data

    paper_scopus_link = None
    for link in paper_entry.get('link', []):
        if link.get('@ref') == 'scopus':
            paper_scopus_link = link.get('@href')
            break

    if paper_scopus_link is not None:

        paper_urls.add(paper_scopus_link)

        try:

            paper_page = _get_paper_page(paper_scopus_link)

            paper_abstract = paper_page.xpath(
                '//section[@id="abstractSection"]//p//text()[normalize-space()]')
            if len(paper_abstract) > 0:
                paper_abstract = re.sub(
                    '\xa0', ' ', ''.join(paper_abstract)).strip()

            authors = paper_page.xpath(
                '//*[@id="authorlist"]/ul/li/span[@class="previewTxt"]')
            paper_authors = []
            for author in authors:
                paper_authors.append(author.text.strip())

            keywords = paper_page.xpath('//*[@id="authorKeywords"]/span')
            for keyword in keywords:
                paper_keywords.add(keyword.text.strip())

            try:
                paper_pages = paper_page.xpath(
                    '//span[@id="journalInfo"]')[0].text.split('Pages')[1].strip()
                if paper_pages.isdigit():  # pragma: no cover
                    paper_number_of_pages = 1
                else:
                    pages_split = paper_pages.split('-')
                    paper_number_of_pages = abs(
                        int(pages_split[0])-int(pages_split[1]))+1
            except Exception:  # pragma: no cover
                pass

        except Exception as e:
            logging.debug(e, exc_info=True)

    paper = Paper(paper_title, paper_abstract, paper_authors, publication,
                  paper_publication_date, paper_urls, paper_doi, paper_citations, paper_keywords,
                  None, paper_number_of_pages, paper_pages)

    return paper


def _get_search_results(search: Search, api_token: str, url: Optional[str] = None) -> dict:  # pragma: no cover
    """
    This method fetch papers from Scopus database using the provided search parameters

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from Scopus database,
    url : Optional[str]
        A predefined URL to be used for the search execution, 
        this is usually used for make the next recursive call on a result pagination
    """

    # is url is not None probably this is a recursive call to the next url of a pagination
    if url is None:
        query = _get_query(search)
        url = f'{BASE_URL}/content/search/scopus?&sort=coverDate&apiKey={api_token}&query={query}'

    headers = {'Accept': 'application/json'}

    return common_util.try_success(lambda: DefaultSession().get(url, headers=headers).json()['search-results'], 2)


def enrich_publication_data(search: Search, api_token: str):
    """
    This method fetch papers from Scopus database to enrich publication data

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from Scopus database,

    Raises
    ------
    AttributeError
        - The API token cannot be null
    """

    if api_token is None or len(api_token.strip()) == 0:
        raise AttributeError('The API token cannot be null')
    
    i = 0
    total = len(search.publication_by_key.items())
    for publication_key, publication in search.publication_by_key.items():
        
        i += 1
        logging.info(f'({i}/{total}) Enriching publication: {publication.title}')

        if publication.issn is not None:

            try:

                publication_entry = _get_publication_entry(
                    publication.issn, api_token)

                if publication_entry is not None:

                    publication_category = publication_entry.get(
                        'prism:aggregationType', None)
                    if publication_category is not None and publication.category is None:
                        publication.category = publication_category

                    publication_publisher = publication_entry.get(
                        'dc:publisher', None)

                    if publication_publisher is not None:
                        publication.publisher = publication_publisher

                    for subject_area in publication_entry.get('subject-area', []):
                        subject_area_value = subject_area.get('$', '').strip()
                        if len(subject_area_value) > 0:
                            publication.subject_areas.add(subject_area_value)

                    publication_cite_score = common_util.try_success(lambda x=publication_entry: float(
                        x.get('citeScoreYearInfoList').get('citeScoreCurrentMetric')))

                    if publication_cite_score is not None:
                        publication.cite_score = publication_cite_score

                    if 'SJRList' in publication_entry and len(publication_entry.get('SJRList').get('SJR')) > 0:
                        publication_sjr = common_util.try_success(lambda x=publication_entry: float(
                            x.get('SJRList').get('SJR')[0].get('$')))

                    if publication_sjr is not None:
                        publication.sjr = publication_sjr

                    if 'SNIPList' in publication_entry and len(publication_entry.get('SNIPList').get('SNIP')) > 0:
                        publication_snip = common_util.try_success(lambda x=publication_entry: float(
                            x.get('SNIPList').get('SNIP')[0].get('$')))

                    if publication_snip is not None:
                        publication.snip = publication_snip

            except Exception:  # pragma: no cover
                pass


def run(search: Search, api_token: str, url: Optional[str] = None, papers_count: Optional[int] = 0):
    """
    This method fetch papers from Scopus database using the provided search parameters
    After fetch the data from Scopus, the collected papers are added to the provided search instance

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from Scopus database,
    url : Optional[str]
        A predefined URL to be used for the search execution, 
        this is usually used for make the next recursive call on a result pagination
    papers_count : Optional[int]
        Papers count used on recursion calls

    Raises
    ------
    AttributeError
        - The API token cannot be null
    """

    if api_token is None or len(api_token.strip()) == 0:
        raise AttributeError('The API token cannot be null')

    search_results = _get_search_results(search, api_token, url)

    total_papers = int(search_results.get('opensearch:totalResults', 0))

    logging.info(f'Scopus: {total_papers} papers to fetch')

    for paper_entry in search_results.get('entry', []):

        if papers_count >= total_papers or search.reached_its_limit(DATABASE_LABEL):
            break

        papers_count += 1

        try:

            paper_title = paper_entry.get("dc:title")
            logging.info(f'({papers_count}/{total_papers}) Fetching Scopus paper: {paper_title}')

            publication = _get_publication(paper_entry, api_token)
            paper = _get_paper(paper_entry, publication)

            if paper is not None:
                paper.add_database(DATABASE_LABEL)
                search.add_paper(paper)

        except Exception as e:  # pragma: no cover
            logging.debug(e, exc_info=True)

    next_url = None
    for link in search_results['link']:
        if link['@ref'] == 'next':
            next_url = link['@href']
            break

    # If there is a next url, the API provided response was paginated and we need to process the next url
    # We'll make a recursive call for it
    if papers_count < total_papers and next_url is not None and not search.reached_its_limit(DATABASE_LABEL):
        run(search, api_token, next_url, papers_count)



import logging
import math
import requests
import datetime
from urllib.parse import urlencode
from typing import Optional
from lxml import html
import findpapers.utils.query_util as query_util
import findpapers.utils.common_util as common_util
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
from findpapers.utils.requests_util import DefaultSession


DATABASE_LABEL = 'ACM'
BASE_URL = 'https://dl.acm.org'
MAX_ENTRIES_PER_PAGE = 100


def _get_search_url(search: Search, start_record: Optional[int] = 0) -> str:
    """
    This method return the URL to be used to retrieve data from ACM database
    See https://dl.acm.org/search/advanced for query tips

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str
        Sequence number of first record to fetch, by default 0

    Returns
    -------
    str
        a URL to be used to retrieve data from ACM database
    """
    
    # when a wildcard is present, the search term cannot be enclosed in quotes
    transformed_query = query_util.replace_search_term_enclosures(search.query, '', '', True)
    
    # some additional query transformations
    transformed_query = transformed_query.replace(' AND NOT ', ' NOT ')
    transformed_query = query_util.replace_search_term_enclosures(transformed_query, '"', '"')

    query = f'Abstract:({transformed_query})'

    # the OR connector between the fields are not working properly, so we'll use only the abstract for now
    #query += f' OR Keyword:({transformed_query})'
    #query += f' OR Title:({transformed_query})'

    url_parameters = {
        'fillQuickSearch': 'false',
        'expand': 'all',
        'AllField': query,
        'pageSize': MAX_ENTRIES_PER_PAGE,
        'startPage': start_record,
        'sortBy': 'Ppub'
    }

    if search.since is not None:
        url_parameters['AfterMonth'] = search.since.month
        url_parameters['AfterYear'] = search.since.year

    if search.until is not None:
        url_parameters['BeforeMonth'] = search.until.month
        url_parameters['BeforeYear'] = search.until.year

    url = f'{BASE_URL}/action/doSearch?{urlencode(url_parameters)}'

    return url


def _get_result(search: Search, start_record: Optional[int] = 0) -> dict:  # pragma: no cover
    """
    This method return results from ACM database using the provided search parameters

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str
        Sequence number of first record to fetch, by default 0

    Returns
    -------
    dict
        a result from ACM database
    """

    url = _get_search_url(search, start_record)

    response = common_util.try_success(lambda: DefaultSession().get(url), 2)
    return html.fromstring(response.content)


def _get_paper_page(url: str) -> html.HtmlElement:  # pragma: no cover
    """
    Get a paper page element from a provided URL

    Parameters
    ----------
    url : str
        The paper URL

    Returns
    -------
    Object
        A HTML element representing the paper given by the provided URL
    """

    response = common_util.try_success(lambda: DefaultSession().get(url), 2)
    return html.fromstring(response.content)


def _get_paper_metadata(doi: str) -> dict:  # pragma: no cover
    """
    Get a paper metadata from a provided DOI

    Parameters
    ----------
    doi : str
        The paper DOI

    Returns
    -------
    dict
        The ACM paper metadata, or None if there's no metadata available
    """

    form = {
        'dois': doi,
        'targetFile': 'custom-bibtex',
        'format': 'bibTex'
    }

    response = common_util.try_success(lambda: DefaultSession().post(
        f'{BASE_URL}/action/exportCiteProcCitation', data=form).json(), 2)

    if response is not None and response.get('items', None) is not None and len(response.get('items')) > 0:
        return response['items'][0][doi]


def _get_paper(paper_page: html.HtmlElement, paper_doi: str, paper_url: str) -> Paper:
    """
    Using a paper entry provided, this method builds a paper instance

    Parameters
    ----------
    paper_page : html.HtmlElement
        A paper page retrieved from ACM
    paper_doi : str
        The paper DOI
    paper_url : str
        The ACM paper URL

    Returns
    -------
    Paper
        A paper instance
    """

    paper_abstract = paper_page.xpath(
        '//*[contains(@class, "abstractSection")]/p')[0].text

    citation_elements = paper_page.xpath(
        '//*[contains(@class, "article-metric citation")]//span')
    paper_citations = None
    if len(citation_elements) == 1:
        paper_citations = int(citation_elements[0].text)

    paper_metadata = _get_paper_metadata(paper_doi)

    if paper_metadata is None:
        return None

    publication = None
    publication_title = paper_metadata.get('container-title', None)

    if publication_title is not None and len(publication_title) > 0:

        publication_isbn = paper_metadata.get('ISBN', None)
        publication_issn = paper_metadata.get('ISSN', None)
        publication_publisher = paper_metadata.get('publisher', None)
        publication_category = paper_metadata.get('type', None)

        publication = Publication(publication_title, publication_isbn,
                                publication_issn, publication_publisher, publication_category)

    paper_title = paper_metadata.get('title', None)

    if paper_title is None or len(paper_title) == 0:
        return None

    paper_authors = paper_metadata.get('author', [])
    paper_authors = ['{} {}'.format(
        x.get('given'), x.get('family')) for x in paper_authors]

    paper_publication_date = None
    if paper_metadata.get('issued', None) != None:
        date_parts = paper_metadata['issued']['date-parts'][0]
        if len(date_parts) == 1:  # only year
            paper_publication_date = datetime.date(date_parts[0], 1, 1)
        else:
            paper_publication_date = datetime.date(
                date_parts[0], date_parts[1], date_parts[2])

    if paper_publication_date is None:
        return None

    paper_keywords = set()
    if paper_metadata.get('keyword', None) is not None:
        paper_keywords = set([x.strip()
                              for x in paper_metadata['keyword'].split(',')])

    paper_pages = paper_metadata.get('page', None)
    if paper_pages is not None:
        paper_pages = paper_pages.replace('\u2013', '-')

    paper_number_of_pages = paper_metadata.get('number-of-pages', None)
    if paper_number_of_pages is not None:
        paper_number_of_pages = int(paper_number_of_pages)

    if paper_doi is None:
        paper_doi = paper_metadata.get('DOI')

    paper = Paper(paper_title, paper_abstract, paper_authors, publication,
                  paper_publication_date, {paper_url}, paper_doi,
                  paper_citations, paper_keywords, None, paper_number_of_pages, paper_pages)

    return paper


def run(search: Search):
    """
    This method fetch papers from ACM database using the provided search parameters
    After fetch the data from ACM, the collected papers are added to the provided search instance

    Parameters
    ----------
    search : Search
        A search instance
    """

    papers_count = 0
    result = _get_result(search)

    try:
        total_papers = int(result.xpath(
            '//*[@class="hitsLength"]')[0].text.strip())
    except Exception:  # pragma: no cover
        total_papers = 0

    logging.info(f'ACM: {total_papers} papers to fetch')

    page_index = 0
    while(papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL)):

        papers_urls = [BASE_URL+x.attrib['href']
                       for x in result.xpath('//*[@class="hlFld-Title"]/a')]

        for paper_url in papers_urls:

            if papers_count >= total_papers or search.reached_its_limit(DATABASE_LABEL):
                break

            try:
                papers_count += 1

                paper_page = _get_paper_page(paper_url)

                paper_title = paper_page.xpath('//*[@class="citation__title"]')[0].text

                logging.info(f'({papers_count}/{total_papers}) Fetching ACM paper: {paper_title}')
                
                paper_doi = None
                if '/abs/' in paper_url:
                    paper_doi = paper_url.split('/abs/')[1]
                elif '/book/' in paper_url:
                    paper_doi = paper_url.split('/book/')[1]
                else:
                    paper_doi = paper_url.split('/doi/')[1]

                paper = _get_paper(paper_page, paper_doi, paper_url)

                if paper is None:
                    continue
                
                paper.add_database(DATABASE_LABEL)

                search.add_paper(paper)

            except Exception as e:  # pragma: no cover
                logging.debug(e, exc_info=True)

        if papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL):
            page_index += 1
            result = _get_result(search, page_index)



import os
import datetime
import logging
import requests
import copy
from lxml import html
from typing import Optional, List
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
import findpapers.searchers.scopus_searcher as scopus_searcher
import findpapers.searchers.ieee_searcher as ieee_searcher
import findpapers.searchers.pubmed_searcher as pubmed_searcher
import findpapers.searchers.arxiv_searcher as arxiv_searcher
import findpapers.searchers.acm_searcher as acm_searcher
import findpapers.utils.common_util as common_util
import findpapers.utils.persistence_util as persistence_util


def _get_paper_metadata_by_url(url: str):
    """
    Private method that returns the paper metadata for a given URL, based on the HTML meta tags

    Parameters
    ----------
    url : str
        A paper URL

    Returns
    -------
    dict
        A paper metadata dict (or None if the paper metadata cannot be found)
    """

    response = common_util.try_success(
        lambda url=url: requests.get(url, allow_redirects=True), 2, 2)

    if response is not None and 'text/html' in response.headers.get('content-type').lower():

        page = html.fromstring(response.content)
        meta_list = page.xpath('//meta')

        paper_metadata = {}

        for meta in meta_list:
            meta_name = meta.attrib.get('name')
            meta_content = meta.attrib.get('content')
            if meta_name is not None and meta_content is not None:

                if meta_name in paper_metadata:
                    if not isinstance(paper_metadata.get(meta_name), list):
                        paper_metadata[meta_name] = [paper_metadata.get(meta_name)]
                    paper_metadata.get(meta_name).append(meta_content)
                else:
                    paper_metadata[meta_name] = meta_content

        return paper_metadata


def _force_single_metadata_value_by_key(metadata_entry: dict, metadata_key: str):
    """
    Sometimes a paper page has some erroneous metadata value duplication, 
    this private method is used to workaround this problem

    Parameters
    ----------
    metadata_entry : dict
        A metadata entry
    metadata_key : str
        A metadata key

    Returns
    -------
    object
        A single value
    """

    return metadata_entry.get(metadata_key, None) if not isinstance(metadata_entry.get(metadata_key, None), list) else metadata_entry.get(metadata_key)[0]



def _enrich(search: Search, scopus_api_token: Optional[str] = None):
    """
    Private method that enriches the search results based on paper metadata

    Parameters
    ----------
    search : Search
        A search instance
    scopus_api_token : Optional[str], optional
        A API token used to fetch data from Scopus database. If you don't have one go to https://dev.elsevier.com and get it, by default None
    """

    i = 0
    total = len(search.papers)
    for paper in list(search.papers):

        i += 1
        logging.info(f'({i}/{total}) Enriching paper: {paper.title}')

        try:

            urls = set()
            if paper.doi is not None:
                urls.add(f'http://doi.org/{paper.doi}')
            else:
                urls = copy.copy(paper.urls)

            for url in urls:

                if 'pdf' in url: # trying to skip PDF links
                    continue

                paper_metadata = _get_paper_metadata_by_url(url)

                if paper_metadata is not None and 'citation_title' in paper_metadata:

                    # when some paper data is present on page's metadata, force to use it. In most of the cases this data is more relyable

                    paper_title = _force_single_metadata_value_by_key(paper_metadata, 'citation_title')
                    
                    if paper_title is None or len(paper_title.strip()) == 0:
                        continue

                    paper.title = paper_title

                    paper_doi = _force_single_metadata_value_by_key(paper_metadata, 'citation_doi')
                    if paper_doi is not None and len(paper_doi.strip()) > 0:
                        paper.doi = paper_doi

                    paper_abstract = _force_single_metadata_value_by_key(paper_metadata, 'citation_abstract')
                    if paper_abstract is not None and len(paper_abstract.strip()) > 0:
                        paper.abstract = paper_abstract
                    
                    paper_authors = paper_metadata.get('citation_author', None)
                    if paper_authors is not None and not isinstance(paper_authors, list): # there is only one author
                        paper_authors = [paper_authors]

                    if paper_authors is not None and len(paper_authors) > 0:
                        paper.authors = paper_authors

                    paper_keywords = _force_single_metadata_value_by_key(paper_metadata, 'keywords')
                    if paper_keywords is not None and len(paper_keywords.strip()) > 0:
                        paper_keywords = set(paper_keywords.split(','))

                    if paper_keywords is not None and len(paper_keywords) > 0:
                        paper.keywords = paper_keywords
                    
                    publication = None
                    publication_title = None
                    publication_category = None
                    if 'citation_journal_title' in paper_metadata:
                        publication_title = _force_single_metadata_value_by_key(paper_metadata, 'citation_journal_title')
                        publication_category = 'Journal'
                    elif 'citation_conference_title' in paper_metadata:
                        publication_title = _force_single_metadata_value_by_key(paper_metadata, 'citation_conference_title')
                        publication_category = 'Conference Proceedings'
                    elif 'citation_book_title' in paper_metadata:
                        publication_title = _force_single_metadata_value_by_key(paper_metadata, 'citation_book_title')
                        publication_category = 'Book'

                    if publication_title is not None and len(publication_title) > 0:
                    
                        publication_issn = _force_single_metadata_value_by_key(paper_metadata, 'citation_issn')
                        publication_isbn = _force_single_metadata_value_by_key(paper_metadata, 'citation_isbn')
                        publication_publisher = _force_single_metadata_value_by_key(paper_metadata, 'citation_publisher')

                        publication = Publication(publication_title, publication_isbn, publication_issn, publication_publisher, publication_category)
                        
                        if paper.publication is None:
                            paper.publication = publication
                        else:
                            paper.publication.enrich(publication)

                    paper_pdf_url = _force_single_metadata_value_by_key(paper_metadata, 'citation_pdf_url')
                    
                    if paper_pdf_url is not None: 
                        paper.add_url(paper_pdf_url)

        except Exception:  # pragma: no cover
            pass

    if scopus_api_token is not None:

        try:
            scopus_searcher.enrich_publication_data(search, scopus_api_token)
        except Exception:  # pragma: no cover
            logging.debug(
                'Error while fetching data from Scopus database', exc_info=True)


def _filter(search: Search):
    """
    Private method that filter the search results

    Parameters
    ----------
    search : Search
        A search instance
    """

    if search.publication_types is not None:
        for paper in list(search.papers):
            try:
                if (paper.publication is not None and paper.publication.category.lower() not in search.publication_types) or \
                    (paper.publication is None and 'other' not in search.publication_types):
                    search.remove_paper(paper)
            except Exception:
                pass


def _database_safe_run(function: callable, search: Search, database_label: str):
    """
    Private method that calls a provided function catching all exceptions without rasing them, only logging a ERROR message

    Parameters
    ----------
    function : callable
        A function that will be call for database fetching
    search : Search
        A search instance
    database_label : str
        A database label
    """
    if not search.reached_its_limit(database_label):
        logging.info(f'Fetching papers from {database_label} database...')
        try:
            function()
        except Exception:  # pragma: no cover
            logging.debug(
                f'Error while fetching papers from {database_label} database', exc_info=True)


def _is_query_ok(query: str) -> bool:
    """
    Checking a search query, it will return True if it's valid or False otherwise.

    Examples
    VALID = ([term a] OR [term b])
    INVALID = ([term a] OR [term b]
    VALID = [term a] OR [term b]
    INVALID = term a OR [term b]
    INVALID = term a [term b]
    INVALID = [] AND [term b]
    VALID = [term a]
    INVALID = []
    INVALID = [

    Parameters
    ----------
    query : str
        A search query to be validated

    Returns
    -------
    bool
        A boolean value indicating whether the query is valid or not
    """

    if len(query) == 0 or len(query) < 3 or query[0] not in ['(', '['] or query[-1] not in [')', ']']:
        return False
    
    # checking groups
    group_characters = []
    for character in query:
        if character in ['(', ')']:
            if len(group_characters) == 0:
                group_characters.append(character)
            else:
                last_group_character = group_characters[-1]

                if last_group_character == character:
                    group_characters.append(character)
                else:
                    group_characters.pop()

    if len(group_characters) > 0: 
        # after the processing above, the list needs to be empty
        return False
    
    # checking keywords and operators
    #TODO: improve this query validation, 'cause this approach ignore the parenthesis
    # and still can return True for invalid queries like [term a] O(R) [term b]

    query_ok = True
    inside_keyword = False
    current_operator = None
    current_keyword = None
    valid_operators = [' AND ', ' OR ', ' AND NOT ']
    transformed_query = query.replace('(','').replace(')','')
    
    for character in transformed_query:

        if inside_keyword:
            if character == ']': # closing a search term
                if current_keyword is None or len(current_keyword.strip()) == 0:
                    query_ok = False
                    break
                current_keyword = None
                inside_keyword = False
            else:
                if current_keyword is None:
                    current_keyword = character
                else:
                    current_keyword += character
        else:
            if character == '[': # opening a search term
                if current_operator is not None and current_operator not in valid_operators:
                    query_ok = False
                    break
                current_operator = None
                inside_keyword = True
            else:
                if current_operator is None:
                    current_operator = character
                else:
                    current_operator += character

    # after the processing above, query_ok needs to be True, 
    # and current_keyword and current_operator need to be null
    return query_ok and current_keyword is None and current_operator is None


def search(outputpath: str, query: Optional[str] = None, since: Optional[datetime.date] = None, until: Optional[datetime.date] = None,
        limit: Optional[int] = None, limit_per_database: Optional[int] = None, databases: Optional[List[str]] = None,
        publication_types: Optional[List[str]] = None, scopus_api_token: Optional[str] = None, ieee_api_token: Optional[str] = None,
        proxy: Optional[str] = None):
    """
    When you have a query and needs to get papers using it, this is the method that you'll need to call.
    This method will find papers from some databases based on the provided query.

    Parameters
    ----------
    outputpath : str
        A valid file path where the search result file will be placed

    query : str, optional

        A query string that will be used to perform the papers search.
        
        If not provided, the query will be loaded from the environment variable FINDPAPERS_QUERY

        All the query terms need to be enclosed in quotes and can be associated using boolean operators,
        and grouped using parentheses. 
        E.g.: [term A] AND ([term B] OR [term C]) AND NOT [term D]

        You can use some wildcards in the query too. Use ? to replace a single character or * to replace any number of characters. 
        E.g.: "son?" -> will match song, sons, ...
        E.g.: "son*" -> will match song, sons, sonar, songwriting, ...

        Note: All boolean operators needs to be uppercased. The boolean operator "NOT" must be preceded by an "AND" operator.

    since : Optional[datetime.date], optional
        A lower bound (inclusive) date that will be used to filter the search results, by default None

    until : Optional[datetime.date], optional
        A upper bound (inclusive) date that will be used to filter the search results, by default None

    limit : Optional[int], optional
        The max number of papers to collect, by default None

    limit_per_database : Optional[int], optional
        The max number of papers to collect per each database, by default None

    databases : List[str], optional
        List of databases where the search should be performed, if not specified all databases will be used, by default None

    publication_types : List[str], optional
        List of publication list of publication types to filter when searching, if not specified all the publication types 
        will be collected (this parameter is case insensitive). The available publication types are: journal, conference proceedings, book, other, by default None

    scopus_api_token : Optional[str], optional
        A API token used to fetch data from Scopus database. If you don't have one go to https://dev.elsevier.com and get it, by default None

    ieee_api_token : Optional[str], optional
        A API token used to fetch data from IEEE database. If you don't have one go to https://developer.ieee.org and get it, by default None
    
    proxy : Optional[str], optional
        proxy URL that can be used during requests. This can be also defined by an environment variable FINDPAPERS_PROXY. By default None

    """

    if proxy is not None:
        os.environ['FINDPAPERS_PROXY'] = proxy
    
    logging.info('Let\'s find some papers, this process may take a while...')

    if databases is not None:
        databases = [x.lower() for x in databases]
    
    if publication_types is not None:
        publication_types = [x.lower().strip() for x in publication_types]
        for publication_type in publication_types:
            if publication_type not in ['journal', 'conference proceedings', 'book', 'other']:
                raise ValueError(f'Invalid publication type: {publication_type}')

    if query is None:
        query = os.getenv('FINDPAPERS_QUERY')

    if query is None or not _is_query_ok(query):
        raise ValueError('Invalid query format')

    common_util.check_write_access(outputpath)

    if ieee_api_token is None:
        ieee_api_token = os.getenv('FINDPAPERS_IEEE_API_TOKEN')

    if scopus_api_token is None:
        scopus_api_token = os.getenv('FINDPAPERS_SCOPUS_API_TOKEN')

    search = Search(query, since, until, limit, limit_per_database, databases=databases, publication_types=publication_types)

    if databases is None or arxiv_searcher.DATABASE_LABEL.lower() in databases:
        _database_safe_run(lambda: arxiv_searcher.run(search),
                        search, arxiv_searcher.DATABASE_LABEL)
    
    if databases is None or pubmed_searcher.DATABASE_LABEL.lower() in databases:
        _database_safe_run(lambda: pubmed_searcher.run(search),
                        search, pubmed_searcher.DATABASE_LABEL)

    if databases is None or acm_searcher.DATABASE_LABEL.lower() in databases:
        _database_safe_run(lambda: acm_searcher.run(search),
                        search, acm_searcher.DATABASE_LABEL)

    if ieee_api_token is not None:
        if databases is None or ieee_searcher.DATABASE_LABEL.lower() in databases:
            _database_safe_run(lambda: ieee_searcher.run(
                search, ieee_api_token), search, ieee_searcher.DATABASE_LABEL)
    else:
        logging.info('IEEE API token not found, skipping search on this database')

    if scopus_api_token is not None:
        if databases is None or scopus_searcher.DATABASE_LABEL.lower() in databases:
            _database_safe_run(lambda: scopus_searcher.run(
                search, scopus_api_token), search, scopus_searcher.DATABASE_LABEL)
    else:
        logging.info('Scopus API token not found, skipping search on this database')

    logging.info('Enriching results...')

    _enrich(search, scopus_api_token)

    logging.info('Filtering results...')

    _filter(search)

    logging.info('Finding and merging duplications...')

    search.merge_duplications()

    logging.info(f'It\'s finally over! {len(search.papers)} papers retrieved. Good luck with your research :)')

    persistence_util.save(search, outputpath)



import os
import logging
from typing import Optional
from findpapers.tools.bibtex_generator_tool import generate_bibtex
from findpapers.tools.search_runner_tool import search
from findpapers.tools.refiner_tool import refine
from findpapers.tools.downloader_tool import download

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)



from __future__ import annotations
import datetime
import re
from typing import List, Set, Optional
from findpapers.models.publication import Publication


class Paper():
    """
    Class that represents a paper instance
    """

    def __init__(self, title: str, abstract: str, authors: List[str], publication: Publication,
                 publication_date: datetime.date, urls: Set[str], doi: Optional[str] = None, citations: Optional[int] = None,
                 keywords: Optional[Set[str]] = None, comments: Optional[str] = None, number_of_pages: Optional[int] = None,
                 pages: Optional[str] = None, databases: Optional[set] = None, selected: Optional[bool] = None,
                 categories: Optional[bool] = None):
        """
        Paper class constructor

        Parameters
        ----------
        title : str
            Paper title
        abstract : str
            Paper abstract
        authors : List[str]
            A list of paper authors
        publication: Publication
            The publication where the paper were published
        publication_date : datetime.date
            Paper publication date
        urls : Set[str]
            Paper urls (one from each scientific database)
        doi : str, optional
            Paper DOI, by default None
        citations : int, optional
            Paper citation count given by scientific databases, by default None
        keywords : Set[str], optional
            Paper keywords, by default None
        comments : str, optional
            Paper comments, by default None
        number_of_pages : int, optional
            Paper number of pages, by default None
        pages : str, optional
            Paper page number or range, by default None
        databases : set, optional
            The databases where the paper was found, by default None
        selected : bool, optional
            If a paper was selected by the user, by default None
        categories : dict, optional
            The defined paper categories by their facets, by default None
        Raises
        ------
        ValueError
            - Paper's title cannot be null
        """

        if title is None or len(title) == 0:
            raise(ValueError('Paper\'s title cannot be null'))

        if publication_date is None:
            raise(ValueError('Paper\'s publication_date cannot be null'))

        self.title = title
        self.abstract = abstract
        self.authors = authors
        self.publication = publication
        self.publication_date = publication_date
        self.urls = urls
        self.doi = doi
        self.citations = citations
        self.keywords = keywords if keywords is not None else set()
        self.comments = comments
        self.number_of_pages = number_of_pages
        self.pages = pages
        self.databases = databases if databases is not None else set()
        self.selected = selected
        self.categories = categories

    def add_database(self, database_name: str):
        """
        Adds database name where the paper was found

        Parameters
        ----------
        database_name : str
            The database name where the paper was found

        Raises
        ------
        ValueError
            - Nowadays only ACM, arXiv, IEEE, PubMed or Scopus are valid database names
        """

        from findpapers.searchers import AVAILABLE_DATABASES

        if database_name not in AVAILABLE_DATABASES:
            raise ValueError(
                f'Invalid database name "{database_name}". Nowadays only {", ".join(AVAILABLE_DATABASES)} are valid database names')

        self.databases.add(database_name)

    def add_url(self, url: str):
        """
        Some paper have URL from many database, this method add this kind of reference to a paper

        Parameters
        ----------
        url : str
            A URL that makes a reference to the paper
        """
        self.urls.add(url)

    def enrich(self, paper: Paper):
        """
        We can enrich some paper information using a duplication of it found in another database.
        This method do this using a provided instance of a duplicated paper

        Parameters
        ----------
        paper : Paper
            A duplication of the "self" paper
        """

        if self.publication_date is None:
            self.publication_date = paper.publication_date

        if self.doi is None:
            self.doi = paper.doi
        
        if self.abstract is None or (paper.abstract is not None and len(paper.abstract) > len(self.abstract)):
            self.abstract = paper.abstract

        if self.authors is None or (paper.authors is not None and len(paper.authors) > len(self.authors)):
            self.authors = paper.authors

        if self.citations is None or (paper.citations is not None and paper.citations > self.citations):
            self.citations = paper.citations

        if self.keywords is None or (paper.keywords is not None and len(paper.keywords) > len(self.keywords)):
            self.keywords = paper.keywords

        if self.comments is None or (paper.comments is not None and len(paper.comments) > len(self.comments)):
            self.comments = paper.comments

        if self.number_of_pages is None or (paper.number_of_pages is not None and paper.number_of_pages > self.number_of_pages):
            self.number_of_pages = paper.number_of_pages

        if self.pages is None or (paper.pages is not None and len(paper.pages) > len(self.pages)):
            self.pages = paper.pages

        for url in paper.urls:
            self.add_url(url)

        for database in paper.databases:
            self.add_database(database)

        if self.publication is None:
            self.publication = paper.publication
        elif paper.publication is not None:
            self.publication.enrich(paper.publication)

    def get_citation_key(self) -> str:
        """
        Get a citation key folowing the pattern <FIRST_AUTHOR><YEAR><TITLE_FIRST_WORD>

        Returns
        -------
        str
            A citation key folowing the pattern <FIRST_AUTHOR><YEAR><TITLE_FIRST_WORD>
        """
        
        author_key = 'unknown'
        if len(self.authors) > 0:
            author_key = self.authors[0].lower().replace(' ', '').replace(',','')
        
        year_key = 'XXXX'
        if self.publication_date is not None:
            year_key = self.publication_date.year
        
        title_key = self.title.split(' ')[0].lower()

        citation_key = re.sub(r'[^\w\d]', '', f'{author_key}{year_key}{title_key}') # keeping only letters, numbers

        return citation_key

    def has_category_match(self, categories: dict) -> bool:
        """
        Check is there is a match with provided categories

        Parameters
        ----------
        categories : dict
            The categories to match

        Returns
        -------
        bool
            if has a match with provided categories
        """

        if categories is not None and self.categories is not None:
            for facet, facet_categories in categories.items():
                for facet_category in facet_categories:
                    if facet_category in self.categories.get(facet, []):
                        return True

        return False

    @classmethod
    def from_dict(cls, paper_dict: dict) -> Paper:
        """
        A method that returns a Paper instance based on the provided dict object

        Parameters
        ----------
        paper_dict : dict
            A dict that represents a Paper instance

        Returns
        -------
        Paper
            A Paper instance based on the provided dict object
        """

        title = paper_dict.get('title')
        abstract = paper_dict.get('abstract')
        authors = paper_dict.get('authors')
        publication = Publication.from_dict(paper_dict.get(
            'publication')) if paper_dict.get('publication') is not None else None
        publication_date = datetime.datetime.strptime(
            paper_dict.get('publication_date'), '%Y-%m-%d').date()
        urls = set(paper_dict.get('urls'))
        doi = paper_dict.get('doi')
        citations = paper_dict.get('citations')
        keywords = set(paper_dict.get('keywords'))
        comments = paper_dict.get('comments')
        number_of_pages = paper_dict.get('number_of_pages')
        pages = paper_dict.get('pages')
        databases = set(paper_dict.get('databases'))
        selected = paper_dict.get('selected')
        categories = paper_dict.get('categories')

        return cls(title, abstract, authors, publication, publication_date, urls, doi, citations, keywords,
                   comments, number_of_pages, pages, databases, selected, categories)

    @staticmethod
    def to_dict(paper: Paper) -> dict:
        """
        A method that returns a dict object based on the provided Paper instance

        Parameters
        ----------
        paper : Paper
            A Paper instance

        Returns
        -------
        dict
            A dict that represents a Paper instance
        """

        return {
            'title': paper.title,
            'abstract': paper.abstract,
            'authors': paper.authors,
            'publication': Publication.to_dict(paper.publication) if paper.publication is not None else None,
            'publication_date': paper.publication_date.strftime('%Y-%m-%d'),
            'urls': list(paper.urls),
            'doi': paper.doi,
            'citations': paper.citations,
            'keywords': list(paper.keywords),
            'comments': paper.comments,
            'number_of_pages': paper.number_of_pages,
            'pages': paper.pages,
            'databases': list(paper.databases),
            'selected': paper.selected,
            'categories': paper.categories,
        }



import requests
import datetime
import logging
import re
import math
import xmltodict
from lxml import html
from typing import Optional
import findpapers.utils.common_util as common_util
import findpapers.utils.query_util as query_util
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
from findpapers.utils.requests_util import DefaultSession


DATABASE_LABEL = 'PubMed'
BASE_URL = 'https://eutils.ncbi.nlm.nih.gov'
MAX_ENTRIES_PER_PAGE = 50


def _get_search_url(search: Search, start_record: Optional[int] = 0) -> str:
    """
    This method return the URL to be used to retrieve data from PubMed database
    See https://www.ncbi.nlm.nih.gov/books/NBK25500/ for query tips

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str, optional
        Sequence number of first record to fetch, by default 0

    Returns
    -------
    str
        a URL to be used to retrieve data from PubMed database
    """
    query = search.query.replace(' AND NOT ', ' NOT ')
    query = query_util.replace_search_term_enclosures(query, '"', '"[TIAB]')

    url = f'{BASE_URL}/entrez/eutils/esearch.fcgi?db=pubmed&term={query} AND has abstract [FILT] AND "journal article"[Publication Type]'

    if search.since is not None or search.until is not None:
        since = datetime.date(
            1, 1, 1) if search.since is None else search.since
        until = datetime.date.today() if search.until is None else search.until

        url += f' AND {since.strftime("%Y/%m/%d")}:{until.strftime("%Y/%m/%d")}[Date - Publication]'

    if start_record is not None:
        url += f'&retstart={start_record}'

    url += f'&retmax={MAX_ENTRIES_PER_PAGE}&sort=pub+date'

    return url


def _get_api_result(search: Search, start_record: Optional[int] = 0) -> dict:  # pragma: no cover
    """
    This method return results from PubMed database using the provided search parameters

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str, optional
        Sequence number of first record to fetch, by default 0

    Returns
    -------
    dict
        a result from PubMed database
    """

    url = _get_search_url(search, start_record)

    return common_util.try_success(lambda: xmltodict.parse(DefaultSession().get(url).content), 2, pre_delay=1)


def _get_paper_entry(pubmed_id: str) -> dict:  # pragma: no cover
    """
    This method return paper data from PubMed database using the provided PubMed ID

    Parameters
    ----------
    pubmed_id : str
        A PubMed ID

    Returns
    -------
    dict
        a paper entry from PubMed database
    """

    url = f'{BASE_URL}/entrez/eutils/efetch.fcgi?db=pubmed&id={pubmed_id}&rettype=abstract'

    return common_util.try_success(lambda: xmltodict.parse(DefaultSession().get(url).content), 2, pre_delay=1)


def _get_publication(paper_entry: dict) -> Publication:
    """
    Using a paper entry provided, this method builds a publication instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from PubMed API

    Returns
    -------
    Publication
        A publication instance
    """

    article = paper_entry.get('PubmedArticleSet').get(
        'PubmedArticle').get('MedlineCitation').get('Article')

    publication_title = article.get('Journal').get('Title')

    if publication_title is None or len(publication_title) == 0:
        return None

    publication_issn = article.get('Journal').get('ISSN').get('#text')

    publication = Publication(publication_title, None,
                              publication_issn, None, 'Journal')

    return publication


def _get_paper(paper_entry: dict, publication: Publication) -> Paper:
    """
    Using a paper entry provided, this method builds a paper instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from IEEE API
    publication : Publication
        A publication instance that will be associated with the paper

    Returns
    -------
    Paper
        A paper instance or None
    """

    article = paper_entry.get('PubmedArticleSet').get(
        'PubmedArticle').get('MedlineCitation').get('Article')

    paper_title = article.get('ArticleTitle', None)

    if paper_title is None or len(paper_title) == 0:
        return None

    paper_title = paper_title if isinstance(paper_title, str) else paper_title.get('#text')

    if 'ArticleDate' in article:
        paper_publication_date_day = article.get('ArticleDate').get('Day')
        paper_publication_date_month = article.get('ArticleDate').get('Month')
        paper_publication_date_year = article.get('ArticleDate').get('Year')
    else:
        paper_publication_date_day = 1
        paper_publication_date_month = common_util.get_numeric_month_by_string(
            article.get('Journal').get('JournalIssue').get('PubDate').get('Month'))
        paper_publication_date_year = article.get('Journal').get(
            'JournalIssue').get('PubDate').get('Year')

    paper_doi = None
    paper_ids = paper_entry.get('PubmedArticleSet').get('PubmedArticle').get(
        'PubmedData').get('ArticleIdList').get('ArticleId')
    for paper_id in paper_ids:
        if paper_id.get('@IdType') == 'doi':
            paper_doi = paper_id.get('#text')
            break

    paper_abstract = None
    paper_abstract_entry = article.get('Abstract', {}).get('AbstractText', None)
    if paper_abstract_entry is None:
        raise ValueError('Paper abstract is empty')

    if isinstance(paper_abstract_entry, list):
        paper_abstract = '\n'.join(
            [x.get('#text') for x in paper_abstract_entry if x.get('#text') is not None])
    else:
        paper_abstract = paper_abstract_entry if isinstance(paper_abstract_entry, str) else paper_abstract_entry.get('#text')

    try:
        paper_keywords = set([x.get('#text') for x in paper_entry.get('PubmedArticleSet').get(
            'PubmedArticle').get('MedlineCitation').get('KeywordList').get('Keyword')])
    except Exception:
        paper_keywords = set()

    try:
        paper_publication_date = datetime.date(int(paper_publication_date_year), int(
            paper_publication_date_month), int(paper_publication_date_day))
    except Exception:
        paper_publication_date = datetime.date(
            int(paper_publication_date_year), 1, 1)

    if paper_publication_date is None:
        return None

    paper_authors = []
    for author in article.get('AuthorList').get('Author'):
        if isinstance(author, str):
            paper_authors.append(author)
        elif isinstance(author, dict):
            paper_authors.append(f"{author.get('ForeName')} {author.get('LastName')}")

    paper_pages = None
    paper_number_of_pages = None
    try:
        paper_pages = article.get('Pagination').get('MedlinePgn')
        if not paper_pages.isdigit(): # if it's a digit, the paper pages range is invalid
            pages_split = paper_pages.split('-')
            paper_number_of_pages = abs(int(pages_split[0])-int(pages_split[1]))+1
    except Exception:  # pragma: no cover
        pass

    paper = Paper(paper_title, paper_abstract, paper_authors, publication,
                  paper_publication_date, set(), paper_doi, None, paper_keywords, None, 
                  paper_number_of_pages, paper_pages)

    return paper


def run(search: Search):
    """
    This method fetch papers from IEEE database using the provided search parameters
    After fetch the data from IEEE, the collected papers are added to the provided search instance

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from IEEE database,

    Raises
    ------
    AttributeError
        - The API token cannot be null
    """

    if search.publication_types is not None and 'journal' not in search.publication_types:
        logging.info('Skiping PubMed search, journal publication type not in filters. Nowadays the PubMed only retrieves papers published on journals.')
        return

    papers_count = 0
    result = _get_api_result(search)

    if result.get('eSearchResult').get('ErrorList', None) is not None:
        total_papers = 0
    else:
        total_papers = int(result.get('eSearchResult').get('Count'))
    
    logging.info(f'PubMed: {total_papers} papers to fetch')

    while(papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL)):

        for pubmed_id in result.get('eSearchResult').get('IdList').get('Id'):

            if papers_count >= total_papers or search.reached_its_limit(DATABASE_LABEL):
                break
            
            papers_count += 1
            
            try:

                paper_entry = _get_paper_entry(pubmed_id)

                if paper_entry is not None:

                    paper_title = paper_entry.get('PubmedArticleSet').get('PubmedArticle').get(
                        'MedlineCitation').get('Article').get('ArticleTitle')

                    logging.info(f'({papers_count}/{total_papers}) Fetching PubMed paper: {paper_title}')

                    publication = _get_publication(paper_entry)
                    paper = _get_paper(paper_entry, publication)

                    if paper is not None:
                        paper.add_database(DATABASE_LABEL)
                        search.add_paper(paper)

            except Exception as e:  # pragma: no cover
                logging.debug(e, exc_info=True)

        if papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL):
            result = _get_api_result(search, papers_count)



import time
import re
import traceback
import logging
import os
import subprocess
import threading
from typing import Optional
from pathlib import Path


def get_numeric_month_by_string(string: str) -> str:
    """
    Get a numeric month representation given a month string representation

    Parameters
    ----------
    string : str
       Month string representation (e.g., jan, january, Jan, Feb, December)

    Returns
    -------
    str
       A month numeric representation (e.g. jan -> 01)
        """

    months = ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
              'jul', 'aug', 'sep', 'oct', 'nov', 'dec']

    return str(months.index(string[:3].lower()) + 1).zfill(2)


def try_success(function, attempts: Optional[int] = 1, pre_delay: Optional[int] = 0, next_try_delay: Optional[int] = 3):
    """
    Try to execute a function and repeat this execution if it raises any exception.
    This function will try N times to succeed, by provided number of attempts.

    Note: you can provide a delay time for pre and post function call,
    so the total delayed time between calls is pre_delay + next_try_delay

    Parameters
    ----------
    function : a common function
            A function that will be tried N times
    attempts : int, optional
            number of attempts, by default 1
    pre_delay : int, optional
            The delay before each function attempts in seconds, by default 0
    next_try_delay : int, optional
            The delay between function attempts in seconds, by default 3

    Returns
    -------
    Object or None
            This method returns the returned value of function or None if function raise Exception in all attempts
    """
    try:
        if attempts > 0:
            time.sleep(pre_delay)
            return function()
        return None
    except Exception as e:
        logging.debug(e, exc_info=True)
        time.sleep(next_try_delay)
        return try_success(function, attempts-1)


def clear(): # pragma: no cover
    """
    Clear the console
    """
    if os.name in ('nt', 'dos'):
        subprocess.call("cls")
    elif os.name in ('linux', 'osx', 'posix'):
        subprocess.call("clear")
    else:
        print('\n') * 120


def check_write_access(path: str):
    """
    Check if you can write on the provided path or raise an error otherwise

    Parameters
    ----------
    path : str
        A OS path

    Raises
    ------
    PermissionError
        If you can't write on the provided path
    """

    try:
        with open(path, 'a'): pass
    except Exception:
        raise PermissionError('You can\'t write on the provided path')


def logging_initialize(verbose: Optional[bool] = False):
    """
    Logging initialize method. If verbose mode is True the logging will be initialized on DEBUG mode. 
    Otherwise, INFO mode will be used

    Parameters
    ----------
    verbose : bool, optional
        If the logging needs to be verbose, by default False
    """

    logging.basicConfig(level=getattr(logging, 'DEBUG' if verbose else 'INFO'),
                        format='%(asctime)s %(levelname)s: %(message)s')


# Based on tornado.ioloop.IOLoop.instance() approach.
# See https://github.com/facebook/tornado
# Whole idea for this metaclass is taken from: https://stackoverflow.com/a/6798042/2402281
class ThreadSafeSingletonMetaclass(type):
    _instances = {}
    _singleton_lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        # double-checked locking pattern (https://en.wikipedia.org/wiki/Double-checked_locking)
        if cls not in cls._instances:
            with cls._singleton_lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(ThreadSafeSingletonMetaclass, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



import findpapers.searchers.scopus_searcher as scopus_searcher
import findpapers.searchers.ieee_searcher as ieee_searcher
import findpapers.searchers.pubmed_searcher as pubmed_searcher
import findpapers.searchers.arxiv_searcher as arxiv_searcher
import findpapers.searchers.acm_searcher as acm_searcher


AVAILABLE_DATABASES = [
    scopus_searcher.DATABASE_LABEL,
    ieee_searcher.DATABASE_LABEL,
    pubmed_searcher.DATABASE_LABEL,
    arxiv_searcher.DATABASE_LABEL,
    acm_searcher.DATABASE_LABEL,
]



import json
import re
from typing import Optional
from findpapers.models.search import Search


def save(search: Search, outputpath: str):
    """
    Method used to save a search result in a JSON representation

    Parameters
    ----------
    search : Search
        A Search instance
    outputpath : str
        A valid file path used to save the search results
    """

    with open(outputpath, 'w') as jsonfile:
        json.dump(Search.to_dict(search), jsonfile, indent=2, sort_keys=True)


def load(search_path: str):
    """
    Method used to load a search result using a JSON representation

    Parameters
    ----------
    search_path : str
        A valid file path containing a JSON representation of the search results
    """

    with open(search_path, 'r') as jsonfile:
        return Search.from_dict(json.load(jsonfile))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['findpapers',
 'findpapers.models',
 'findpapers.searchers',
 'findpapers.tools',
 'findpapers.utils']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.3,<0.5.0',
 'edlib>=1.3.8,<2.0.0',
 'inquirer>=2.7.0,<3.0.0',
 'lxml>=4.5.2,<5.0.0',
 'requests>=2.24.0,<3.0.0',
 'typer>=0.3.2,<0.4.0',
 'xmltodict>=0.12.0,<0.13.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['findpapers = findpapers.cli:main']}

setup_kwargs = {
    'name': 'findpapers',
    'version': '0.4.3',
    'description': 'Findpapers is an application that helps researchers who are looking for references for their research.',
    'long_description': '# Findpapers\n\n[![PyPI - License](https://img.shields.io/pypi/l/findpapers)](https://gitlab.com/jonatasgrosman/findpapers/-/blob/master/LICENSE)\n[![PyPI](https://img.shields.io/pypi/v/findpapers)](https://pypi.org/project/findpapers)\n[![pipeline status](https://gitlab.com/jonatasgrosman/findpapers/badges/master/pipeline.svg)](https://gitlab.com/jonatasgrosman/findpapers/-/commits/master)\n<!--[![coverage report](https://gitlab.com/jonatasgrosman/findpapers/badges/master/coverage.svg)](https://gitlab.com/jonatasgrosman/findpapers/-/commits/master)-->\n\n\n\nFindpapers is an application that helps researchers who are looking for references for their research. The application will perform searches in several databases (currently ACM, arXiv, IEEE, PubMed, and Scopus) from a user-defined search query.\n\nIn summary, this tool will help you to perform the process below:\n\n![Workflow](docs/workflow.png)\n\n# Requirements\n\n- Python 3.7+\n\n# Installation\n\n```console\n$ pip install findpapers\n```\n\nYou can check your findpapers version running:\n\n```console\n$ findpapers version\n```\n\nIf you have an old version of the tool and want to upgrade it run the following command:\n\n```console\n$ pip install findpapers --upgrade\n```\n\n# How to use it?\n\nAll application actions are command-line based. The available commands are \n\n- ```findpapers search```: Search for papers metadata using a query. This search will be made by matching the query with the paper\'s title, abstract, and keywords.\n\n- ```findpapers refine```: Refine the search results by selecting/classifying the papers\n\n- ```findpapers download```: Download full-text papers using the search results\n\n- ```findpapers bibtex```: Generate a BibTeX file from the search results\n\nYou can control the commands logging verbosity by the **-v** (or **--verbose**) argument.\n\nIn the following sections, we will show how to use the findpapers commands. However, all the commands have the **--help** argument to display some summary about their usage, E.g., ```findpapers search --help```.\n\n## Search query construction\n\nFirst of all, we need to know how to build the search queries. The search queries must follow the rules:\n\n- All the query terms need to be not empty and enclosed by square brackets. E.g., **[term a]**\n\n- The query can contain boolean operators, but they must be uppercase. The allowed operators are AND, OR, and NOT. E.g., **[term a] AND [term b]**\n\n- All the operators must have one and only one whitespace before and after them. E.g., **[term a] OR [term b] OR [term c]**\n\n- The NOT operator must always be preceded by an AND operator E.g., **[term a] AND NOT [term b]**\n\n- A subquery needs to be enclosed by parentheses. E.g., **[term a] AND ([term b] OR [term c])**\n\n- The composition of terms is only allowed through boolean operators. Queries like "**[term a] [term b]**" are invalid\n\nYou can use some wildcards in the query too. Use question mark (?) to replace exactly one character, and use an asterisk (*) to replace zero or more characters:\n\n- **[son?]** will match song, sons, ... (But won\'t match "son")\n\n- **[son\\*]** will match son, song, sons, sonic, songwriting, ...\n\nThere are some rules that you\'ll need to follow when using wildcards:\n\n- Cannot be used at the start of a search term;\n- A minimum of 3 characters preceding the asterisk wildcard (*) is required;\n- The asterisk wildcard (*) can only be used at the end of a search term;\n- Can be used only in single terms;\n- Only one wildcard can be included in a search term.\n\nNote: The IEEE and PubMed databases don\'t support the "?" wildcard.\n\nLet\'s see some examples of valid and invalid queries:\n\n| Query  | Valid? |\n| ------------- | ------------- |\n| [term a]   |  Yes  |\n| ([term a] OR [term b])   |  Yes  |\n| [term a] OR [term b]  |  Yes  |\n| [term a] AND [term b]   |  Yes  |\n| [term a] AND NOT ([term b] OR [term c])  |  Yes  |\n| [term a] OR ([term b] AND ([term\\*] OR [t?rm]))  |  Yes |\n| [term a]OR[term b]   |  **No** (no whitespace between terms and boolean operator)  |\n| [term a] &nbsp;&nbsp;OR&nbsp;&nbsp; [term b]  |  **No** (more than 1 whitespace between terms and boolean operator)  |\n| ([term a] OR [term b]  |  **No** (missing parentheses)  |\n| [term a] or [term b]  |  **No** (lowercase boolean operator)  |\n| term a OR [term b]  |  **No** (missing square brackets)  |\n| [term a] [term b]  |  **No** (missing boolean operator)  |\n| [term a] XOR [term b] |  **No** (invalid boolean operator)   |\n| [term a] OR NOT [term b] |  **No** (NOT boolean operator must be preceded by AND)   |\n| [] AND [term b]  |  **No** (empty term)  |\n|[some term\\*]  |  **No** (wildcards can be used only in single terms)  |\n|[?erm]  |  **No** (wildcards cannot be used at the start of a search term)  |\n|[te*]  |  **No** (a minimum of 3 characters preceding the asterisk wildcard is required)  |\n|[ter*s]  |  **No** (the asterisk wildcard can only be used at the end of a search term)  |\n|[t?rm?]  |  **No** (only one wildcard can be included in a search term)  |\n\n## Basic example (TL;DR)\n\n- Searching for papers:\n\n```console\n$ findpapers search /some/path/search.json -q "[happiness] AND ([joy] OR [peace of mind]) AND NOT [stressful]"\n```\n\n- Refining search results:\n\n```console\n$ findpapers refine /some/path/search.json\n```\n\n- Downloading full-text from selected papers:\n\n```console\n$ findpapers download /some/path/search.json /some/path/papers/ -s\n```\n\n- Generating BibTeX file from selected papers:\n\n```console\n$ findpapers bibtex /some/path/search.json /some/path/mybib.bib -s\n```\n\n## Advanced example\n\nThis advanced usage documentation can be a bit boring to read (and write), so I think it\'s better to go for a storytelling approach here.\n\n*Let\'s take a look at Dr. McCartney\'s research. He\'s a computer scientist interested in AI and music, so he created a search query to collect papers that can help with his research and exported this query to an environment variable.*\n\n```console\n$ export QUERY="([artificial intelligence] OR [AI] OR [machine learning] OR [ML] OR [deep learning] OR [DL]) AND ([music] OR [s?ng])"\n```\n\n*Dr. McCartney is interested in testing his query, so he decides to collect only 20 papers to test whether the query is suitable for his research (the Findpapers results are sorted by publication date in descending order).*\n\n```console\n$ findpapers search /some/path/search_paul.json --query "$QUERY" --limit 20\n```\n\n*But after taking a look at the results contained in the ```search_paul.json``` file, he notices two problems:*\n - *Only one database was used to collect all the 20 papers*\n - *Some collected papers were about drums, but he doesn\'t like drums or drummers*\n\n*So he decides to solve these problems by:*\n- *Reformulating his query, and also placing it inside a file to make his life easier.*\n\n```/some/path/query.txt```\n```\n([artificial intelligence] OR [AI] OR [machine learning] OR [ML] OR [deep learning] OR [DL]) AND ([music] OR [s?ng]) AND NOT [drum*]\n```\n\n- *Performing the search limiting the number of papers that can be collected by each database.*\n\n```console\n$ findpapers search /some/path/search_paul.json --query-file /some/path/query.txt --limit-db 4\n```\n\n*Now his query returned the papers he wanted, but he realized one thing, no papers were collected from Scopus or IEEE databases. Then he noticed that he needed to pass his Scopus and IEEE API access keys when calling the search command. So he went to https://dev.elsevier.com and https://developer.ieee.org, generated the access keys, and used them in the search.*\n\n```console\n$ export IEEE_TOKEN=SOME_SUPER_SECRET_TOKEN\n\n$ export SCOPUS_TOKEN=SOME_SUPER_SECRET_TOKEN\n\n$ findpapers search /some/path/search_paul.json --query-file /some/path/query.txt --limit-db 4 --token-ieee "$IEEE_TOKEN" --token-scopus "$SCOPUS_TOKEN"\n```\n\n*Now everything is working as he expected, so it\'s time to do the final papers search. So he defines that he wants to collect only works published between 2000 and 2020. He also decides that he only wants papers collected from ACM, IEEE, and Scopus. And he only wants to papers published on a journal or conference proceedings (Tip: The available publication types on Findpapers are: journal, conference proceedings, book, other. When a particular publication does not fit into any of the other types it is classified as "other", e.g., magazines, newsletters, unpublished manuscripts)*\n\n```console\n$ findpapers search /some/path/search_paul.json --query-file /some/path/query.txt --token-ieee "$IEEE_TOKEN" --token-scopus "$SCOPUS_TOKEN" --since 2000-01-01 --until 2020-12-31 --databases "acm,ieee,scopus" --publication-types "journal,conference proceedings"\n```\n\n*The searching process took a long time, but after many cups of coffee, Dr. McCartney finally has a good list of papers with the potential to help in his research. All the information collected is in the ```search_paul.json``` file. He could access this file now and manually filter which works are most interesting for him, but he prefers to use the Findpapers ```refine``` command for this.*\n\n*First, he wants to filter the papers looking only at their basic information.*\n\n```console\n$ findpapers refine /some/path/search_paul.json\n```\n\n![refine-01](docs/refine-01.jpeg)\n\n*After completing the first round filtering of the collected papers, he decides to do new filtering on the selected ones looking at the paper\'s extra info (citations, DOI, publication name, etc.) and abstract now. He also chooses to perform some classification while doing this further filtering (tip: he\'ll need to use spacebar for categories selection). And to help in this process, he also decides to highlight some keywords contained in the abstract.*\n\n```console\n$ export CATEGORIES_CONTRIBUTION="Contribution:Metric,Tool,Model,Method"\n\n$ export CATEGORIES_RESEARCH_TYPE="Research Type:Validation Research,Solution Proposal,Philosophical,Opinion,Experience,Other"\n\n$ export HIGHLIGHTS="propose, achiev, accuracy, method, metric, result, limitation, state of the art"\n\n$ findpapers refine /some/path/search_paul.json --selected --abstract --extra-info --categories "$CATEGORIES_CONTRIBUTION" --categories "$CATEGORIES_RESEARCH_TYPE" --highlights "$HIGHLIGHTS"\n```\n\n![refine-02](docs/refine-02.jpeg)\n\nAn interesting point to stand out from the tool is that it automatically prevents duplication of papers, merging their information when the same paper is found in different databases. You can see this in the image above, where the Findpapers found the same work on the IEEE and Scopus databases (see "Paper found in" value) and merged the paper information on a single record.\n\n\n*Now that Dr. McCartney has selected all the papers he wanted, he wants to see all of them.*\n\n```console\n$ findpapers refine /some/path/search_paul.json --selected --abstract --extra-info --list\n```\n\n*He wants to see all the removed papers too.*\n\n```console\n$ findpapers refine /some/path/search_paul.json --removed --abstract --extra-info --list\n```\n\n*Then, he decides to download the full-text from all the selected papers which have a "Model" or "Tool" as a contribution.*\n\n```console\n$ findpapers download /some/path/search_paul.json /some/path/papers --selected --categories "Contribution:Tool,Model"\n```\n\n*He also wants to generate the BibTeX file from these papers.*\n\n```console\n$ findpapers bibtex /some/path/search_paul.json /some/path/mybib.bib --selected --categories "Contribution:Tool,Model"\n```\n\n*But when he compared the papers\' data in the ```/some/path/mybib.bib```  and PDF files in the ```/some/path/papers``` folder, he noticed that many papers had not been downloaded.*\n\n*So when he opened the ```/some/path/papers/download.log``` file, he could see the URL of all papers that weren\'t downloaded correctly. After accessing these links, he noticed that some of them weren\'t downloaded due to some limitations of Findpapers (currently, the tool has a set of heuristics to perform the download that may not work in all cases). However, the vast majority of papers weren\'t downloaded because they were behind a paywall. But, Dr. McCartney has access to these papers when he\'s connected to the network at the university where he works, but unfortunately, he is at home right now.*\n\n*But he discovers two things that could save him from this mess. First, the university provides a proxy for tunneling requests. Second, Findpapers accepts the configuration of a proxy URL. And of course, he\'ll use this feature (see the "--proxy" argument on the command bellow).*\n\n```console\n$ findpapers download /some/path/search_paul.json /some/path/papers --selected --categories "Contribution:Tool,Model" --proxy "https://mccartney:super_secret_pass@liverpool.ac.uk:1234"\n```\n\n*Now the vast majority of the papers he has access have been downloaded correctly.*\n\n*And at the end of it, he decides to download the full-text from all the selected works (regardless of their categorization) and generate their BibTeX file too.*\n\n```console\n$ findpapers download /some/path/search_paul.json /some/path/papers --selected\n\n$ findpapers bibtex /some/path/search_paul.json /some/path/mybib.bib --selected\n```\n\n*That\'s all, folks! We have reached the end of our journey. I hope Dr. McCartney can continue his research and publish his work without any major problems now.*\n\nAs you could see, all the information collected and enriched by the Findpapers is placed in a single JSON file. From this file, it is possible to create interesting visualizations about the collected data ...\n\n![charts](docs/charts.png)\n\n... so, use your imagination! (the above visualization was made by the [samples/charts.py](https://gitlab.com/jonatasgrosman/findpapers/-/blob/master/samples/charts.py) script)\n\nWith the story above, we cover all the commands available in Findpapers. I know this documentation is unconventional, but I haven\'t had time to write a more formal version of the documentation. But you can help us to improve this, take a look at the next section and see how you can do that.\n\n\n# Want to help?\n\nSee the [contribution guidelines](https://gitlab.com/jonatasgrosman/findpapers/-/blob/master/CONTRIBUTING.md)\nif you\'d like to contribute to Findpapers project.\n\nYou don\'t even need to know how to code to contribute to the project. Even the improvement of our documentation is an outstanding contribution.\n\nIf this project has been useful for you, please share it with your friends. This project could be helpful for them too.\n\nIf you like this project and want to motivate the maintainers, give us a :star:. This kind of recognition will make us very happy with the work that we\'ve done :heart:\n\n\n---\n\n**Note**: If you\'re seen this project from GitHub, this is just a mirror, \nthe official project source code is hosted [here](https://gitlab.com/jonatasgrosman/findpapers) on GitLab.',
    'author': 'Jonatas Grosman',
    'author_email': 'jonatasgrosman@gmail.com',
    'maintainer': 'Jonatas Grosman',
    'maintainer_email': 'jonatasgrosman@gmail.com',
    'url': 'https://gitlab.com/jonatasgrosman/findpapers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import requests
import datetime
import logging
import re
import math
from lxml import html
from typing import Optional
import findpapers.utils.common_util as common_util
import findpapers.utils.query_util as query_util
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
from findpapers.utils.requests_util import DefaultSession


DATABASE_LABEL = 'IEEE'
BASE_URL = 'http://ieeexploreapi.ieee.org'
MAX_ENTRIES_PER_PAGE = 200


def _get_search_url(search: Search, api_token: str, start_record: Optional[int] = 1) -> str:
    """
    This method return the URL to be used to retrieve data from IEEE database
    See https://developer.ieee.org/docs/read/Metadata_API_details for query tips

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from IEEE database,
    start_record : str
        Sequence number of first record to fetch, by default 1

    Returns
    -------
    str
        a URL to be used to retrieve data from IEEE database
    """

    query = search.query.replace(' AND NOT ', ' NOT ')
    query = query_util.replace_search_term_enclosures(query, '"Abstract":"', '"')

    url = f'{BASE_URL}/api/v1/search/articles?querytext=({query})&format=json&apikey={api_token}&max_records={MAX_ENTRIES_PER_PAGE}'

    if search.since is not None:
        url += f'&start_year={search.since.year}'

    if search.until is not None:
        url += f'&end_year={search.until.year}'

    if start_record is not None:
        url += f'&start_record={start_record}'

    if search.publication_types is not None:

        content_types = set()

        if 'conference proceedings' in search.publication_types:
            content_types.add('Conferences')
        if 'journal' in search.publication_types:
            content_types.add('Journals')
        if 'book' in search.publication_types:
            content_types.add('Books')
        if 'other' in search.publication_types:
            content_types.add('Courses')
            content_types.add('Early Access')
            content_types.add('Magazines')
            content_types.add('Standards')
            
        url += f'&content_type={",".join(content_types)}'

    return url


def _get_api_result(search: Search, api_token: str, start_record: Optional[int] = 1) -> dict:  # pragma: no cover
    """
    This method return results from IEEE database using the provided search parameters

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from IEEE database,
    start_record : str
        Sequence number of first record to fetch, by default 1

    Returns
    -------
    dict
        a result from IEEE database
    """

    url = _get_search_url(search, api_token, start_record)

    return common_util.try_success(lambda: DefaultSession().get(url).json(), 2)


def _get_publication(paper_entry: dict) -> Publication:
    """
    Using a paper entry provided, this method builds a publication instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from IEEE API

    Returns
    -------
    Publication
        A publication instance or None
    """

    publication_title = paper_entry.get('publication_title', None)

    if publication_title is None or len(publication_title) == 0:
        return None

    publication_isbn = paper_entry.get('isbn', None)
    publication_issn = paper_entry.get('issn', None)
    publication_publisher = paper_entry.get('publisher', None)
    publication_category = paper_entry.get('content_type', None)

    publication = Publication(publication_title, publication_isbn,
                              publication_issn, publication_publisher, publication_category)

    return publication


def _get_paper(paper_entry: dict, publication: Publication) -> Paper:
    """
    Using a paper entry provided, this method builds a paper instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from IEEE API
    publication : Publication
        A publication instance that will be associated with the paper

    Returns
    -------
    Paper
        A paper instance or None
    """

    paper_title = paper_entry.get('title', None)

    if paper_title is None or len(paper_title) == 0:
        return None

    paper_publication_date = paper_entry.get('publication_date', None)
    paper_doi = paper_entry.get('doi', None)
    paper_citations = paper_entry.get('citing_paper_count', None)
    paper_abstract = paper_entry.get('abstract', None)
    paper_urls = {paper_entry.get('pdf_url')}
    paper_pages = None
    paper_number_of_pages = None

    try:
        paper_keywords = set(paper_entry.get(
            'index_terms').get('author_terms').get('terms'))
    except Exception as e:
        paper_keywords = set()

    if paper_publication_date is not None:
        try:
            paper_publication_date_split = paper_publication_date.split(' ')
            day = int(paper_publication_date_split[0].split('-')[0])
            month = int(common_util.get_numeric_month_by_string(
                paper_publication_date_split[1]))
            year = int(paper_publication_date_split[2])

            paper_publication_date = datetime.date(year, month, day)
        except Exception as e:
            pass

    if not isinstance(paper_publication_date, datetime.date):
        paper_publication_date = datetime.date(
            paper_entry.get('publication_year'), 1, 1)

    if paper_publication_date is None:
        return None

    paper_authors = []
    for author in paper_entry.get('authors').get('authors'):
        paper_authors.append(author.get('full_name'))

    start_page = paper_entry.get('start_page', None)
    end_page = paper_entry.get('end_page', None)
    

    if start_page is not None and end_page is not None:
        try:
            paper_pages = f"{paper_entry.get('start_page')}-{paper_entry.get('end_page')}"
            paper_number_of_pages = abs(
                int(paper_entry.get('start_page'))-int(paper_entry.get('end_page')))+1
        except Exception:  # pragma: no cover
            pass

    paper = Paper(paper_title, paper_abstract, paper_authors, publication,
                  paper_publication_date, paper_urls, paper_doi, paper_citations, 
                  paper_keywords, None, paper_number_of_pages, paper_pages)

    return paper


def run(search: Search, api_token: str):
    """
    This method fetch papers from IEEE database using the provided search parameters
    After fetch the data from IEEE, the collected papers are added to the provided search instance

    Parameters
    ----------
    search : Search
        A search instance
    api_token : str
        The API key used to fetch data from IEEE database,

    Raises
    ------
    AttributeError
        - The API token cannot be null
    """

    if api_token is None or len(api_token.strip()) == 0:
        raise AttributeError('The API token cannot be null')

    papers_count = 0
    result = _get_api_result(search, api_token)
    total_papers = result.get('total_records')

    logging.info(f'IEEE: {total_papers} papers to fetch')

    while(papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL)):

        for paper_entry in result.get('articles'):

            if papers_count >= total_papers or search.reached_its_limit(DATABASE_LABEL):
                break
            
            papers_count += 1

            try:

                logging.info(f'({papers_count}/{total_papers}) Fetching IEEE paper: {paper_entry.get("title")}')

                publication = _get_publication(paper_entry)
                paper = _get_paper(paper_entry, publication)

                if paper is not None:
                    paper.add_database(DATABASE_LABEL)
                    search.add_paper(paper)

            except Exception as e:  # pragma: no cover
                logging.debug(e, exc_info=True)

        if papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL):
            result = _get_api_result(search, api_token, papers_count+1)



import requests
import datetime
import logging
import re
import math
import xmltodict
from lxml import html
from typing import Optional
import findpapers.utils.common_util as common_util
import findpapers.utils.query_util as query_util
from findpapers.models.search import Search
from findpapers.models.paper import Paper
from findpapers.models.publication import Publication
from findpapers.utils.requests_util import DefaultSession


DATABASE_LABEL = 'arXiv'
BASE_URL = 'http://export.arxiv.org'
MAX_ENTRIES_PER_PAGE = 200
SUBJECT_AREA_BY_KEY = {
    'astro-ph': 'Astrophysics',
    'astro-ph.CO': 'Cosmology and Nongalactic Astrophysics',
    'astro-ph.EP': 'Earth and Planetary Astrophysics',
    'astro-ph.GA': 'Astrophysics of Galaxies',
    'astro-ph.HE': 'High Energy Astrophysical Phenomena',
    'astro-ph.IM': 'Instrumentation and Methods for Astrophysics',
    'astro-ph.SR': 'Solar and Stellar Astrophysics',
    'cond-mat.dis-nn': 'Disordered Systems and Neural Networks',
    'cond-mat.mes-hall': 'Mesoscale and Nanoscale Physics',
    'cond-mat.mtrl-sci': 'Materials Science',
    'cond-mat.other': 'Other Condensed Matter',
    'cond-mat.quant-gas': 'Quantum Gases',
    'cond-mat.soft': 'Soft Condensed Matter',
    'cond-mat.stat-mech': 'Statistical Mechanics',
    'cond-mat.str-el': 'Strongly Correlated Electrons',
    'cond-mat.supr-con': 'Superconductivity',
    'cs.AI': 'Artificial Intelligence',
    'cs.AR': 'Hardware Architecture',
    'cs.CC': 'Computational Complexity',
    'cs.CE': 'Computational Engineering, Finance, and Science',
    'cs.CG': 'Computational Geometry',
    'cs.CL': 'Computation and Language',
    'cs.CR': 'Cryptography and Security',
    'cs.CV': 'Computer Vision and Pattern Recognition',
    'cs.CY': 'Computers and Society',
    'cs.DB': 'Databases',
    'cs.DC': 'Distributed, Parallel, and Cluster Computing',
    'cs.DL': 'Digital Libraries',
    'cs.DM': 'Discrete Mathematics',
    'cs.DS': 'Data Structures and Algorithms',
    'cs.ET': 'Emerging Technologies',
    'cs.FL': 'Formal Languages and Automata Theory',
    'cs.GL': 'General Literature',
    'cs.GR': 'Graphics',
    'cs.GT': 'Computer Science and Game Theory',
    'cs.HC': 'Human-Computer Interaction',
    'cs.IR': 'Information Retrieval',
    'cs.IT': 'Information Theory',
    'cs.LG': 'Learning',
    'cs.LO': 'Logic in Computer Science',
    'cs.MA': 'Multiagent Systems',
    'cs.MM': 'Multimedia',
    'cs.MS': 'Mathematical Software',
    'cs.NA': 'Numerical Analysis',
    'cs.NE': 'Neural and Evolutionary Computing',
    'cs.NI': 'Networking and Internet Architecture',
    'cs.OH': 'Other Computer Science',
    'cs.OS': 'Operating Systems',
    'cs.PF': 'Performance',
    'cs.PL': 'Programming Languages',
    'cs.RO': 'Robotics',
    'cs.SC': 'Symbolic Computation',
    'cs.SD': 'Sound',
    'cs.SE': 'Software Engineering',
    'cs.SI': 'Social and Information Networks',
    'cs.SY': 'Systems and Control',
    'econ.EM': 'Econometrics',
    'eess.AS': 'Audio and Speech Processing',
    'eess.IV': 'Image and Video Processing',
    'eess.SP': 'Signal Processing',
    'gr-qc': 'General Relativity and Quantum Cosmology',
    'hep-ex': 'High Energy Physics - Experiment',
    'hep-lat': 'High Energy Physics - Lattice',
    'hep-ph': 'High Energy Physics - Phenomenology',
    'hep-th': 'High Energy Physics - Theory',
    'math.AC': 'Commutative Algebra',
    'math.AG': 'Algebraic Geometry',
    'math.AP': 'Analysis of PDEs',
    'math.AT': 'Algebraic Topology',
    'math.CA': 'Classical Analysis and ODEs',
    'math.CO': 'Combinatorics',
    'math.CT': 'Category Theory',
    'math.CV': 'Complex Variables',
    'math.DG': 'Differential Geometry',
    'math.DS': 'Dynamical Systems',
    'math.FA': 'Functional Analysis',
    'math.GM': 'General Mathematics',
    'math.GN': 'General Topology',
    'math.GR': 'Group Theory',
    'math.GT': 'Geometric Topology',
    'math.HO': 'History and Overview',
    'math.IT': 'Information Theory',
    'math.KT': 'K-Theory and Homology',
    'math.LO': 'Logic',
    'math.MG': 'Metric Geometry',
    'math.MP': 'Mathematical Physics',
    'math.NA': 'Numerical Analysis',
    'math.NT': 'Number Theory',
    'math.OA': 'Operator Algebras',
    'math.OC': 'Optimization and Control',
    'math.PR': 'Probability',
    'math.QA': 'Quantum Algebra',
    'math.RA': 'Rings and Algebras',
    'math.RT': 'Representation Theory',
    'math.SG': 'Symplectic Geometry',
    'math.SP': 'Spectral Theory',
    'math.ST': 'Statistics Theory',
    'math-ph': 'Mathematical Physics',
    'nlin.AO': 'Adaptation and Self-Organizing Systems',
    'nlin.CD': 'Chaotic Dynamics',
    'nlin.CG': 'Cellular Automata and Lattice Gases',
    'nlin.PS': 'Pattern Formation and Solitons',
    'nlin.SI': 'Exactly Solvable and Integrable Systems',
    'nucl-ex': 'Nuclear Experiment',
    'nucl-th': 'Nuclear Theory',
    'physics.acc-ph': 'Accelerator Physics',
    'physics.ao-ph': 'Atmospheric and Oceanic Physics',
    'physics.app-ph': 'Applied Physics',
    'physics.atm-clus': 'Atomic and Molecular Clusters',
    'physics.atom-ph': 'Atomic Physics',
    'physics.bio-ph': 'Biological Physics',
    'physics.chem-ph': 'Chemical Physics',
    'physics.class-ph': 'Classical Physics',
    'physics.comp-ph': 'Computational Physics',
    'physics.data-an': 'Data Analysis, Statistics and Probability',
    'physics.ed-ph': 'Physics Education',
    'physics.flu-dyn': 'Fluid Dynamics',
    'physics.gen-ph': 'General Physics',
    'physics.geo-ph': 'Geophysics',
    'physics.hist-ph': 'History and Philosophy of Physics',
    'physics.ins-det': 'Instrumentation and Detectors',
    'physics.med-ph': 'Medical Physics',
    'physics.optics': 'Optics',
    'physics.plasm-ph': 'Plasma Physics',
    'physics.pop-ph': 'Popular Physics',
    'physics.soc-ph': 'Physics and Society',
    'physics.space-ph': 'Space Physics',
    'q-bio.BM': 'Biomolecules',
    'q-bio.CB': 'Cell Behavior',
    'q-bio.GN': 'Genomics',
    'q-bio.MN': 'Molecular Networks',
    'q-bio.NC': 'Neurons and Cognition',
    'q-bio.OT': 'Other Quantitative Biology',
    'q-bio.PE': 'Populations and Evolution',
    'q-bio.QM': 'Quantitative Methods',
    'q-bio.SC': 'Subcellular Processes',
    'q-bio.TO': 'Tissues and Organs',
    'q-fin.CP': 'Computational Finance',
    'q-fin.EC': 'Economics',
    'q-fin.GN': 'General Finance',
    'q-fin.MF': 'Mathematical Finance',
    'q-fin.PM': 'Portfolio Management',
    'q-fin.PR': 'Pricing of Securities',
    'q-fin.RM': 'Risk Management',
    'q-fin.ST': 'Statistical Finance',
    'q-fin.TR': 'Trading and Market Microstructure',
    'quant-ph': 'Quantum Physics',
    'stat.AP': 'Applications',
    'stat.CO': 'Computation',
    'stat.ME': 'Methodology',
    'stat.ML': 'Machine Learning',
    'stat.OT': 'Other Statistics',
    'stat.TH': 'Statistics Theory'
}


def _get_search_url(search: Search, start_record: Optional[int] = 0) -> str:
    """
    This method return the URL to be used to retrieve data from arXiv database
    See https://arxiv.org/help/api/user-manual for query tips

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str
        Sequence number of first record to fetch, by default 0

    Returns
    -------
    str
        a URL to be used to retrieve data from arXiv database
    """

    transformed_query = search.query.replace(' AND NOT ', ' ANDNOT ')
    if transformed_query[0] == '"':
        transformed_query = ' ' + transformed_query
    transformed_query = transformed_query.replace(' [', ' FIELD_TYPE:[')
    transformed_query = transformed_query.replace('([', '(FIELD_TYPE:[')

    # when a wildcard is present, the search term cannot be enclosed in quotes
    transformed_query = query_util.replace_search_term_enclosures(transformed_query, '', '', True)
    transformed_query = query_util.replace_search_term_enclosures(transformed_query, '"', '"').strip()

    abstract_query = transformed_query.replace('FIELD_TYPE:', 'abs:')
    title_query = transformed_query.replace('FIELD_TYPE:', 'ti:')
    final_query = f'({title_query}) OR ({abstract_query})'

    url = f'{BASE_URL}/api/query?search_query={final_query}&start={start_record}&sortBy=submittedDate&sortOrder=descending&max_results={MAX_ENTRIES_PER_PAGE}'

    return url


def _get_api_result(search: Search, start_record: Optional[int] = 0) -> dict: # pragma: no cover
    """
    This method return results from arXiv database using the provided search parameters

    Parameters
    ----------
    search : Search
        A search instance
    start_record : str
        Sequence number of first record to fetch, by default 1

    Returns
    -------
    dict
        a result from arXiv database
    """

    url = _get_search_url(search, start_record)

    return common_util.try_success(lambda: xmltodict.parse(DefaultSession().get(url).content), 2, pre_delay=1)


def _get_publication(paper_entry: dict) -> Publication:
    """
    Using a paper entry provided, this method builds a publication instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from arXiv API

    Returns
    -------
    Publication, or None
        A publication instance
    """

    if 'arxiv:journal_ref' in paper_entry:

        publication_title = paper_entry.get('arxiv:journal_ref').get('#text')

        if publication_title is None or len(publication_title) == 0:
            return None

        subject_areas = set()

        if 'category' in paper_entry:
            if isinstance(paper_entry.get('category'), list):
                for category in paper_entry.get('category'):
                    subject_area = SUBJECT_AREA_BY_KEY.get(category.get('@term'), None)
                    if subject_area is not None:
                        subject_areas.add(subject_area)
            else:
                subject_area = SUBJECT_AREA_BY_KEY.get(paper_entry.get('category').get('@term'), None)
                if subject_area is not None:
                    subject_areas.add(subject_area)

        publication = Publication(
            publication_title, subject_areas=subject_areas)

        return publication


def _get_paper(paper_entry: dict, paper_publication_date: datetime.date, publication: Publication) -> Paper:
    """
    Using a paper entry provided, this method builds a paper instance

    Parameters
    ----------
    paper_entry : dict
        A paper entry retrieved from arXiv API
    paper_publication_date : datetime.date
        The paper publication date
    publication : Publication
        A publication instance that will be associated with the paper

    Returns
    -------
    Paper
        A paper instance
    """

    paper_title = paper_entry.get('title', None)

    if paper_title is None or len(paper_title) == 0:
        return None

    paper_title = paper_title.replace('\n','') 
    paper_title = re.sub(' +', ' ', paper_title)

    paper_doi = paper_entry.get('arxiv:doi').get(
        '#text') if 'arxiv:doi' in paper_entry else None
    paper_abstract = paper_entry.get('summary', None)
    paper_urls = set()
    paper_authors = []

    if 'link' in paper_entry:
        if isinstance(paper_entry.get('link'), list):
            for link in paper_entry.get('link'):
                paper_urls.add(link.get('@href'))
        else:
            paper_urls.add(paper_entry.get('link').get('@href'))

    if 'author' in paper_entry:
        if isinstance(paper_entry.get('author'), list):
            for author in paper_entry.get('author'):
                paper_authors.append(author.get('name'))
        else:
            paper_authors.append(paper_entry.get('author').get('name'))

    paper_comments = paper_entry.get('arxiv:comment', {}).get('#text', None)

    paper = Paper(paper_title, paper_abstract, paper_authors, publication,
                  paper_publication_date, paper_urls, paper_doi, comments=paper_comments)

    return paper


def run(search: Search):
    """
    This method fetch papers from arXiv database using the provided search parameters
    After fetch the data from arXiv, the collected papers are added to the provided search instance

    Parameters
    ----------
    search : Search
        A search instance

    """

    papers_count = 0
    result = _get_api_result(search)

    total_papers = int(result.get('feed').get(
        'opensearch:totalResults').get('#text'))

    logging.info(f'arXiv: {total_papers} papers to fetch')

    while(papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL)):

        for paper_entry in result.get('feed').get('entry'):

            if papers_count >= total_papers or search.reached_its_limit(DATABASE_LABEL):
                break

            papers_count += 1

            try:

                paper_title = paper_entry.get("title")
                logging.info(f'({papers_count}/{total_papers}) Fetching arXiv paper: {paper_title}')

                published_date = datetime.datetime.strptime(
                    paper_entry.get('published')[:10], '%Y-%m-%d').date()

                # nowadays we don't have a date filter on arXiv API, so we need to do it by ourselves'
                if search.since is not None and published_date < search.since:
                    logging.info(
                        'Skipping paper due to "since" date constraint')
                    continue
                elif search.until is not None and published_date > search.until:
                    logging.info(
                        'Skipping paper due to "until" date constraint')
                    continue

                publication = _get_publication(paper_entry)
                paper = _get_paper(paper_entry, published_date, publication)

                if paper is not None:
                    paper.add_database(DATABASE_LABEL)
                    search.add_paper(paper)

            except Exception as e:  # pragma: no cover
                logging.debug(e, exc_info=True)

        if papers_count < total_papers and not search.reached_its_limit(DATABASE_LABEL):
            result=_get_api_result(search, papers_count)
